"""Docstring in public module."""
import os
import sys

import ujson as json
from tornado.testing import AsyncHTTPTestCase

APP_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.append(os.path.join(APP_ROOT, ".."))


class TestUserProfile(AsyncHTTPTestCase):
    def get_app(self):
        from consoleme.routes import make_app

        return make_app(jwt_validator=lambda x: {})

    def test_profile(self):
        from consoleme.config import config

        self.maxDiff = None
        headers = {
            config.get("auth.user_header_name"): "user@example.com",
            config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }

        response = self.fetch("/api/v2/user_profile", headers=headers)
        self.assertEqual(response.code, 200)
        response_j = json.loads(response.body)
        consoleme_logo = response_j["site_config"].pop("consoleme_logo")
        self.assertIn("/images/logos/", consoleme_logo)
        self.assertEqual(
            response_j,
            {
                "site_config": {
                    "google_analytics": {"tracking_id": None, "options": {}},
                    "documentation_url": "https://github.com/Netflix/consoleme/",
                    "support_contact": "consoleme-support@example.com",
                    "support_chat_url": "https://www.example.com/slack/channel",
                    "security_logo": None,
                    "security_url": None,
                    "landing_url": None,
                    "notifications": {"enabled": None, "request_interval": 60},
                    "cloudtrail_denies_policy_generation": True,
                },
                "user": "user@example.com",
                "can_logout": False,
                "is_contractor": False,
                "employee_photo_url": "https://www.gravatar.com/avatar/b58996c504c5638798eb6b511e6f49af?d=mp",
                "employee_info_url": None,
                "authorization": {
                    "can_edit_policies": False,
                    "can_create_roles": False,
                    "can_delete_iam_principals": False,
                },
                "pages": {
                    "header": {
                        "custom_header_message_title": "Warning:",
                        "custom_header_message_text": "This configuration has no authentication configured. Do not use in a production environment. Do not publicly expose this endpoint or your AWS environment will be compromised! Click [here](/generate_config) to generate a configuration.",
                        "custom_header_message_route": ".*",
                    },
                    "groups": {"enabled": False},
                    "users": {"enabled": False},
                    "policies": {"enabled": True},
                    "self_service": {"enabled": True},
                    "api_health": {"enabled": False},
                    "audit": {"enabled": False},
                    "config": {"enabled": False},
                },
                "accounts": {"123456789012": "default_account"},
            },
        )



import asyncio
import sys
from datetime import datetime, timedelta

import pytz
import sentry_sdk
from email_validator import validate_email

from consoleme.config import config
from consoleme.handlers.base import BaseAPIV2Handler, TornadoRequestHandler
from consoleme.lib.auth import can_admin_all
from consoleme.lib.dynamo import UserDynamoHandler
from consoleme.lib.jwt import generate_jwt_token
from consoleme.lib.password import check_password_strength
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.lib.web import handle_generic_error_response
from consoleme.models import (
    AuthenticationResponse,
    LoginAttemptModel,
    RegistrationAttemptModel,
    UserManagementModel,
    WebResponse,
)

stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
log = config.get_logger()


class UserRegistrationHandler(TornadoRequestHandler):
    """
    Allows user registration if it is configured.
    """

    def initialize(self):
        self.ddb = UserDynamoHandler()

    async def post(self):
        # TODO: Send verification e-mail to proposed user

        log_data = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "message": "Attempting to register user",
            "user-agent": self.request.headers.get("User-Agent"),
        }

        generic_error_message: str = "User registration failed"
        # Fail if getting users by password is not enabled
        if not config.get("auth.get_user_by_password"):
            errors = [
                "Expected configuration `auth.get_user_by_password`, but it is not enabled."
            ]
            await handle_generic_error_response(
                self, generic_error_message, errors, 403, "not_configured", log_data
            )
            return
        # Fail if user registration not allowed
        if not config.get("auth.allow_user_registration"):
            errors = [
                "Expected configuration `auth.allow_user_registration`, but it is not enabled."
            ]
            await handle_generic_error_response(
                self, generic_error_message, errors, 403, "not_configured", log_data
            )
            return

        registration_attempt = RegistrationAttemptModel.parse_raw(self.request.body)
        log_data["username"] = registration_attempt.username
        # Fail if username not valid email address
        try:
            if not validate_email(registration_attempt.username):
                errors = ["Username must be a valid e-mail address."]
                await handle_generic_error_response(
                    self,
                    generic_error_message,
                    errors,
                    403,
                    "invalid_request",
                    log_data,
                )
                return
        except Exception as e:
            sentry_sdk.capture_exception()
            await handle_generic_error_response(
                self, generic_error_message, [str(e)], 403, "invalid_request", log_data
            )
            return
        # Fail if user already exists
        if await self.ddb.get_user(registration_attempt.username):
            errors = ["User already exists"]
            await handle_generic_error_response(
                self, generic_error_message, errors, 403, "invalid_request", log_data
            )
            return

        # Fails if password is not strong enough.
        password_strength_errors = await check_password_strength(
            registration_attempt.password
        )
        if password_strength_errors:
            await handle_generic_error_response(
                self,
                password_strength_errors["message"],
                password_strength_errors["errors"],
                403,
                "weak_password",
                log_data,
            )
            return

        self.ddb.create_user(
            registration_attempt.username, registration_attempt.password
        )

        res = WebResponse(
            status="success",
            status_code=200,
            message=f"Successfully created user {registration_attempt.username}.",
        )
        self.write(res.json(exclude_unset=True))


class LoginConfigurationHandler(TornadoRequestHandler):
    def get(self):
        default_configuration = {
            "enabled": config.get("auth.get_user_by_password"),
            "page_title": config.get(
                "LoginConfigurationHandler.page_title",
                "Welcome to ConsoleMe - Please Sign-In",
            ),
            "allow_password_login": config.get("auth.get_user_by_password", True),
            "allow_sso_login": config.get(
                "LoginConfigurationHandler.allow_sso_login", True
            ),
            "allow_sign_up": config.get("auth.allow_user_registrationp", False),
            "custom_message": "",
        }
        login_configuration = config.get(
            "LoginConfigurationHandler.login_configuration", default_configuration
        )
        self.write(login_configuration)


class LoginHandler(TornadoRequestHandler):
    """
    Handles user log-in flow if password authentication is enabled.
    """

    def initialize(self):
        self.ddb = UserDynamoHandler()

    def check_xsrf_cookie(self):
        pass

    def set_default_headers(self) -> None:
        self.set_header("Content-Type", "application/json")

    async def post(self):
        log_data = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "message": "Attempting to authenticate User",
            "user-agent": self.request.headers.get("User-Agent"),
        }
        generic_error_message = "Authentication failed"
        if not config.get("auth.get_user_by_password"):
            errors = [
                "Expected configuration `auth.get_user_by_password`, but it is not enabled."
            ]
            await handle_generic_error_response(
                self, generic_error_message, errors, 403, "not_configured", log_data
            )
            return
        # Auth cookie must be set to use password authentication.
        if not config.get("auth.set_auth_cookie"):
            errors = [
                "Expected configuration `auth.set_auth_cookie`, but it is not enabled."
            ]
            await handle_generic_error_response(
                self, generic_error_message, errors, 403, "not_configured", log_data
            )
            return

        login_attempt = LoginAttemptModel.parse_raw(self.request.body)
        log_data["username"] = login_attempt.username
        log_data["after_redirect_uri"] = login_attempt.after_redirect_uri
        authenticated_response: AuthenticationResponse = (
            await self.ddb.authenticate_user(login_attempt)
        )
        if not authenticated_response.authenticated:
            # Wait 1 second to protect from single-host brute-force
            await asyncio.sleep(1)
            await handle_generic_error_response(
                self,
                generic_error_message,
                authenticated_response.errors,
                403,
                "authentication_failure",
                log_data,
            )
            return
        # Make and set jwt for user
        expiration = datetime.utcnow().replace(tzinfo=pytz.UTC) + timedelta(
            minutes=config.get("jwt.expiration_minutes", 60)
        )
        encoded_cookie = await generate_jwt_token(
            authenticated_response.username,
            authenticated_response.groups,
            exp=expiration,
        )
        self.set_cookie(
            config.get("auth_cookie_name", "consoleme_auth"),
            encoded_cookie,
            expires=expiration,
            secure=config.get(
                "auth.cookie.secure",
                True if "https://" in config.get("url") else False,
            ),
            httponly=config.get("auth.cookie.httponly", True),
            samesite=config.get("auth.cookie.samesite", True),
        )
        res = WebResponse(
            status="redirect",
            redirect_url=login_attempt.after_redirect_uri,
            status_code=200,
            reason="authenticated_redirect",
            message="User has successfully authenticated. Redirecting to their intended destination.",
        )
        self.write(res.json(exclude_unset=True, exclude_none=True))


class UserManagementHandler(BaseAPIV2Handler):
    """
    Handles creating and updating users. Only authorized users are allowed to access this endpoint.
    """

    def initialize(self):
        self.ddb = UserDynamoHandler()

    async def post(self):
        log_data = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "user": self.user,
            "message": "Create/Update User",
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
            "ip": self.ip,
        }
        generic_error_message = "Unable to create/update user"
        log.debug(log_data)
        # Checks authz levels of current user
        if not can_admin_all(self.user, self.groups):
            errors = ["User is not authorized to access this endpoint."]
            await handle_generic_error_response(
                self, generic_error_message, errors, 403, "unauthorized", log_data
            )
            return
        request = UserManagementModel.parse_raw(self.request.body)
        log_data["requested_user"] = request.username
        if request.user_management_action.value == "create":
            log.debug(
                {
                    **log_data,
                    "message": "Creating user",
                    "requested_user": request.username,
                    "requested_groups": request.groups,
                }
            )

            # Fails if password is not strong enough.
            password_strength_errors = await check_password_strength(request.password)
            if password_strength_errors:
                await handle_generic_error_response(
                    self,
                    password_strength_errors["message"],
                    password_strength_errors["errors"],
                    403,
                    "weak_password",
                    log_data,
                )
                return

            self.ddb.create_user(
                request.username,
                request.password,
                request.groups,
            )
            res = WebResponse(
                status="success",
                status_code=200,
                message=f"Successfully created user {request.username}.",
            )
            self.write(res.json(exclude_unset=True, exclude_none=True))
            return
        elif request.user_management_action.value == "update":
            log.debug(
                {
                    **log_data,
                    "message": "Updating user",
                    "requested_user": request.username,
                    "requested_groups": request.groups,
                }
            )

            if request.password:
                # Fails if password is not strong enough.
                password_strength_errors = await check_password_strength(
                    request.password
                )
                if password_strength_errors:
                    await handle_generic_error_response(
                        self,
                        password_strength_errors["message"],
                        password_strength_errors["errors"],
                        403,
                        "weak_password",
                        log_data,
                    )
                    return
            self.ddb.update_user(
                request.username,
                request.password,
                request.groups,
            )
            res = WebResponse(
                status="success",
                status_code=200,
                message=f"Successfully updated user {request.username}.",
            )
            self.write(res.json(exclude_unset=True, exclude_none=True))
            return
        elif request.user_management_action.value == "delete":
            log.debug(
                {
                    **log_data,
                    "message": "Deleting user",
                    "requested_user": request.username,
                }
            )
            self.ddb.delete_user(
                request.username,
            )
            res = WebResponse(
                status="success",
                status_code=200,
                message=f"Successfully deleted user {request.username}.",
            )
            self.write(res.json(exclude_unset=True, exclude_none=True))
            return
        else:
            errors = ["Change type is not supported by this endpoint."]
            await handle_generic_error_response(
                self, generic_error_message, errors, 403, "invalid_request", log_data
            )
            return



"""Docstring in public module."""
import os
import sys

import ujson as json
from tornado.testing import AsyncHTTPTestCase

APP_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.append(os.path.join(APP_ROOT, ".."))


class TestRoleLoginApi(AsyncHTTPTestCase):
    def get_app(self):
        from consoleme.routes import make_app

        return make_app(jwt_validator=lambda x: {})

    def test_role_api_fail(self):
        from consoleme.config import config

        headers = {
            config.get("auth.user_header_name"): "user@example.com",
            config.get("auth.groups_header_name"): "groupa,groupb,groupc",
        }

        response = self.fetch("/api/v2/role_login/role123", headers=headers)
        self.assertEqual(response.code, 404)
        self.assertEqual(
            json.loads(response.body),
            {
                "type": "error",
                "message": "You do not have any roles matching your search criteria. ",
            },
        )

    def test_role_api_fail_multiple_matching_roles(self):
        from consoleme.config import config

        headers = {
            config.get("auth.user_header_name"): "userwithmultipleroles@example.com",
            config.get("auth.groups_header_name"): "group9,group3",
        }

        response = self.fetch("/api/v2/role_login/role", headers=headers)
        self.assertEqual(response.code, 200)
        response_j = json.loads(response.body)
        self.assertEqual(
            response_j["message"],
            "You have more than one role matching your query. Please select one.",
        )
        self.assertEqual(response_j["reason"], "multiple_roles")
        self.assertEqual(response_j["type"], "redirect")
        self.assertIn("/?arn=role&warningMessage=", response_j["redirect_url"])

    def test_role_api_success(self):
        from consoleme.config import config

        headers = {
            config.get("auth.user_header_name"): "userwithrole@example.com",
            config.get("auth.groups_header_name"): "groupa@example.com",
        }

        response = self.fetch("/api/v2/role_login/roleA", headers=headers)
        self.assertEqual(response.code, 200)
        response_j = json.loads(response.body)
        self.assertEqual(response_j["type"], "redirect")
        self.assertEqual(response_j["reason"], "console_login")
        self.assertEqual(response_j["role"], "arn:aws:iam::123456789012:role/roleA")
        self.assertIn(
            "https://signin.aws.amazon.com/federation?Action=login&Issuer=YourCompany&Destination=https%3A%2F%2Fus-east-1.console.aws.amazon.com&SigninToken=",
            response_j["redirect_url"],
        )



import asyncio
import os
from typing import List

from asgiref.sync import sync_to_async


async def bound_fetch(sem, fn, args, kwargs):
    # Getter function with semaphore.
    async with sem:
        return {
            "fn": fn,
            "args": args,
            "kwargs": kwargs,
            "result": await fn(*args, **kwargs),
        }


async def bound_fetch_sync(sem, fn, args, kwargs):
    # Getter function with semaphore.
    async with sem:
        return {
            "fn": fn,
            "args": args,
            "kwargs": kwargs,
            "result": await sync_to_async(fn)(*args, **kwargs),
        }


async def run_in_parallel(task_list: List, threads=os.cpu_count(), sync=True):
    async def run():
        sem = asyncio.Semaphore(threads)
        futures = []
        for task in task_list:
            if sync:
                futures.append(
                    asyncio.ensure_future(
                        bound_fetch_sync(
                            sem,
                            task.get("fn"),
                            task.get("args", ()),
                            task.get("kwargs", {}),
                        )
                    )
                )
            else:
                futures.append(
                    asyncio.ensure_future(
                        bound_fetch(
                            sem,
                            task.get("fn"),
                            task.get("args", ()),
                            task.get("kwargs", {}),
                        )
                    )
                )
        responses = asyncio.gather(*futures)
        return await responses

    return await run()



"""
Defining constants and resolving variables through configuration file
"""

import yaml

config_yaml = yaml.load(open("config.yaml"), Loader=yaml.FullLoader)

domain_prefix = config_yaml["domain_prefix"]

BASE_NAME = "ConsoleMeECS"
SPOKE_BASE_NAME = "ConsoleMeSpoke"
MAIN_ACCOUNT_ID = config_yaml["main_account"]

APPLICATION_PREFIX = "consoleme-" + domain_prefix
APPLICATION_SUFFIX = "secure".lower()
HOSTED_ZONE_ID = config_yaml["hosted_zone_id"]
HOSTED_ZONE_NAME = config_yaml["hosted_zone_name"]
ADMIN_TEMP_PASSWORD = config_yaml["admin_temp_password"]
USE_PUBLIC_DOCKER_IMAGE = config_yaml["use_public_docker_image"]
DOCKER_IMAGE = "consoleme/consoleme"
MIN_CAPACITY = config_yaml["min_capacity"]
MAX_CAPACITY = config_yaml["max_capacity"]
CONFIG_SECRET_NAME = "ConsoleMeConfigSecret"



import json
import sys
from hashlib import sha256

import sentry_sdk
import tornado.escape
import tornado.web

from consoleme.celery_tasks.celery_tasks import app as celery_app
from consoleme.config import config
from consoleme.handlers.base import BaseHandler
from consoleme.lib.auth import can_edit_dynamic_config
from consoleme.lib.dynamo import UserDynamoHandler
from consoleme.lib.json_encoder import SetEncoder
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.lib.redis import RedisHandler

ddb = UserDynamoHandler()
log = config.get_logger()
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
aws = get_plugin_by_name(config.get("plugins.aws", "default_aws"))()
red = RedisHandler().redis_sync()


class DynamicConfigApiHandler(BaseHandler):
    def on_finish(self) -> None:
        if self.request.method != "POST":
            return
        # Force a refresh of credential authorization mapping in current region
        # TODO: Trigger this to run cross-region
        # TODO: Delete server-side user-role cache intelligently so users get immediate access
        celery_app.send_task(
            "consoleme.celery_tasks.celery_tasks.cache_credential_authorization_mapping",
            countdown=config.get("dynamic_config.dynamo_load_interval"),
        )

    async def get(self) -> None:
        """
        Get the dynamic configuration endpoint.
        ---
        description: Presents a YAML-configured editor to allow viewing and modification of dynamic config
        responses:
            200:
                description: View of dynamic configuration
            403:
                description: Unauthorized to access this page
        """

        if not self.user:
            return

        if not can_edit_dynamic_config(self.user, self.groups):
            raise tornado.web.HTTPError(
                403, "Only application admins are authorized to view this page."
            )

        dynamic_config = await ddb.get_dynamic_config_yaml()

        self.write(
            {
                "dynamicConfig": dynamic_config.decode("utf-8"),
                "sha256": sha256(dynamic_config).hexdigest(),
            }
        )

    async def post(self):
        """
        Post an update to the dynamic configuration endpoint.
        ---
        description: Update dynamic configuration
        responses:
            200:
                description: Update successful.
            403:
                description: Unauthorized to access this page
        """

        if not self.user:
            return
        if not can_edit_dynamic_config(self.user, self.groups):
            raise tornado.web.HTTPError(
                403, "Only application admins are authorized to view this page."
            )
        existing_dynamic_config = await ddb.get_dynamic_config_yaml()
        if existing_dynamic_config:
            existing_dynamic_config_sha256 = sha256(existing_dynamic_config).hexdigest()
        else:
            existing_dynamic_config_sha256 = None
        result = {"status": "success"}
        log_data = {
            "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
            "user": self.user,
            "existing_dynamic_config_sha256": existing_dynamic_config_sha256,
        }
        log.debug(log_data)

        data = tornado.escape.json_decode(self.request.body)
        try:
            existing_sha256 = data.get("existing_sha256")
            new_sha256 = sha256(data["new_config"].encode("utf-8")).hexdigest()
            if existing_sha256 == new_sha256:
                raise Exception(
                    "You didn't change the dynamic configuration. Try again!"
                )
            if (
                existing_dynamic_config_sha256
                and not existing_dynamic_config_sha256 == existing_sha256
            ):
                raise Exception(
                    "Dynamic configuration was updated by another user before your changes were processed. "
                    "Please refresh your page and try again."
                )

            await ddb.update_dynamic_config(data["new_config"], self.user)
        except Exception as e:
            result["status"] = "error"
            result["error"] = f"There was an error processing your request: {e}"
            sentry_sdk.capture_exception()
            self.write(json.dumps(result, cls=SetEncoder))
            await self.finish()
            return

        result["newConfig"] = data["new_config"]
        result["newsha56"] = new_sha256
        self.write(result)
        await self.finish()
        # Force a refresh of dynamic configuration in the current region. Other regions will need to wait until the
        # next background thread refreshes it automatically. By default, this happens every 60 seconds.
        config.CONFIG.load_config_from_dynamo(ddb=ddb, red=red)
        return



import base64
import json as original_json
import time
from collections import defaultdict

import simplejson as json
from boto3.dynamodb.types import Binary  # noqa

from consoleme.config import config
from consoleme.lib.aws import get_iam_principal_owner, simulate_iam_principal_action
from consoleme.lib.cache import store_json_results_in_redis_and_s3
from consoleme.lib.dynamo import UserDynamoHandler
from consoleme.lib.json_encoder import SetEncoder
from consoleme.lib.notifications.models import (
    ConsoleMeUserNotification,
    ConsoleMeUserNotificationAction,
)

ddb = UserDynamoHandler()


class CloudTrail:
    async def process_cloudtrail_errors(
        self,
        aws,
        notification_ttl_seconds=config.get(
            "process_cloudtrail_errors.notification_ttl", 86400
        ),
    ) -> object:
        """
        Processes Cloudtrail Errors that were cached by the `cache_cloudtrail_denies` celery task. Generates and returns
        count data. If configured, generates notifications to end-users based on policies that can be generated

        :param notification_ttl_seconds:
        :return:
        """
        notification_type = "cloudtrail_generated_policy"
        expiration = int(time.time() + notification_ttl_seconds)

        # Get all existing cloudtrail_generated_policy notifications
        all_notifications = {}
        all_notifications_l = await ddb.parallel_scan_table_async(
            ddb.notifications_table
        )
        for existing_notification in all_notifications_l:
            if existing_notification["type"] != notification_type:
                continue
            all_notifications[
                existing_notification["predictable_id"]
            ] = ConsoleMeUserNotification.parse_obj(existing_notification)

        error_count = {}
        # Get all existing Cloudtrail errors. This will be expensive if there are a large number of errors.
        cloudtrail_errors = await ddb.parallel_scan_table_async(
            ddb.cloudtrail_table,
        )
        cloudtrail_errors = ddb._data_from_dynamo_replace(cloudtrail_errors)
        error_count = ddb.count_arn_errors(error_count, cloudtrail_errors)
        new_or_changed_notifications = {}
        for cloudtrail_error in cloudtrail_errors:
            arn = cloudtrail_error.get("arn", "")
            principal_owner = await get_iam_principal_owner(arn, aws)
            session_name = cloudtrail_error.get("session_name", "")
            principal_type = "iam" + arn.split(":")[-1].split("/")[0]
            account_id = arn.split(":")[4]
            principal_name = arn.split("/")[-1]
            url_role_path = (
                f"/policies/edit/{account_id}/{principal_type}/{principal_name}"
            )
            event_call = cloudtrail_error.get("event_call", "")
            resource = cloudtrail_error.get("resource", "")
            # If a given IAM principal encounters an sts:AssumeRole AccessDeny error for a given role across multiple
            # accounts (I.E. The role being assumed has the same name on multiple accounts), we only want to generate
            # one notification for the issue. Therefore, resource_full_name might not accurately reflect the resource
            # name, but it works for the purposes of creating a unique ID such that only one notification gets created
            # for this set of CloudTrail Events
            resource_full_name = resource.split(":")[-1]
            predictable_id = f"{notification_type}-{principal_owner}-{principal_name}-{event_call}-{resource_full_name}"
            generated_request = {
                "role": {
                    "name": principal_name,
                    "account_id": account_id,
                    "account_name": "",
                    "arn": arn,
                    "cloudtrail_details": {
                        "error_url": "",
                        "errors": {
                            "cloudtrail_errors": [],
                        },
                    },
                    "s3_details": {
                        "error_url": "",
                        "errors": {
                            "s3_errors": [],
                        },
                    },
                    "apps": {
                        "app_details": [],
                    },
                    "tags": [],
                    "templated": False,
                    "principal": {
                        "principal_type": "AwsResource",
                        "principal_arn": arn,
                    },
                },
                "updated_policy": json.dumps(cloudtrail_error["generated_policy"]),
            }
            encoded_request = base64.b64encode(
                json.dumps(generated_request).encode()
            ).decode("utf-8")
            encoded_request_url = f"/selfservice?encoded_request={encoded_request}"
            notification_message = config.get(
                "process_cloudtrail_errors.generate_notifications.message",
                """We've generated a policy suggestion for a recent permissions error with **[{arn}]({url_role_path})**.
Please click the button below to review it.

You are receiving this notification because your team owns this role, or you were using this role at the time the error
was detected. This notification will disappear when a similar error has not occurred for 24 hours.""".format(
                    arn=arn,
                    url_role_path=url_role_path,
                ),
            )

            message_actions = [
                ConsoleMeUserNotificationAction.parse_obj(
                    {
                        "http_method": "get",
                        "uri": encoded_request_url,
                        "text": "Review and Submit Request",
                    }
                )
            ]

            generated_notification = ConsoleMeUserNotification(
                predictable_id=predictable_id,
                type=notification_type,
                users_or_groups=set(),
                event_time=cloudtrail_error["epoch_event_time"],
                expiration=expiration,
                expired=False,
                message=notification_message,
                message_actions=message_actions,
                details=cloudtrail_error,
                read_by_users=set(),
                read_by_all=False,
                hidden_for_users=set(),
                hidden_for_all=False,
                version=1,
            )

            # Update existing item without overwriting settings
            if all_notifications.get(predictable_id):
                new_or_changed_notifications[predictable_id] = all_notifications[
                    predictable_id
                ]
                new_or_changed_notifications[
                    predictable_id
                ].event_time = cloudtrail_error["epoch_event_time"]
                new_or_changed_notifications[predictable_id].expiration = expiration
                new_or_changed_notifications[predictable_id].expired = False
                new_or_changed_notifications[
                    predictable_id
                ].message = notification_message
                new_or_changed_notifications[predictable_id].details = cloudtrail_error
            else:
                # Maybe Add IAM policy simulation result to the CloudTrail event, and in turn, the notification details
                # We don't want to simulate events for every single update of a notification, just one time for
                # Initial creation
                if config.get(
                    "process_cloudtrail_errors.simulate_iam_principal_action"
                ):
                    generated_notification.details[
                        "iam_policy_simulation"
                    ] = await simulate_iam_principal_action(
                        arn, event_call, resource, cloudtrail_error.get("source_ip")
                    )
            if principal_owner and not all_notifications.get(predictable_id):
                generated_notification.users_or_groups.add(principal_owner)
                new_or_changed_notifications[predictable_id] = generated_notification

            # Also link Notifications with users via session nme
            if not session_name or session_name.startswith(
                "i-"
            ):  # Session ID is instance ID
                continue

            if new_or_changed_notifications.get(predictable_id):
                new_or_changed_notifications[predictable_id].users_or_groups.add(
                    session_name
                )
            else:
                new_or_changed_notifications[predictable_id] = generated_notification
                new_or_changed_notifications[predictable_id].users_or_groups.add(
                    session_name
                )

            # Optionally add development users to the notification
            new_or_changed_notifications[predictable_id].users_or_groups.update(
                set(config.get("process_cloudtrail_errors.additional_notify_users", []))
            )
        new_or_changed_notifications_l = []
        notifications_by_user_group = defaultdict(list)
        for notification in new_or_changed_notifications.values():
            new_or_changed_notifications_l.append(notification.dict())
            for user_or_group in notification.users_or_groups:
                notifications_by_user_group[user_or_group].append(notification.dict())
        if new_or_changed_notifications_l:
            ddb.parallel_write_table(
                ddb.notifications_table, new_or_changed_notifications_l
            )
        if notifications_by_user_group:
            for k, v in notifications_by_user_group.items():
                notifications_by_user_group[k] = original_json.dumps(v, cls=SetEncoder)
            await store_json_results_in_redis_and_s3(
                notifications_by_user_group,
                redis_key=config.get("notifications.redis_key", "ALL_NOTIFICATIONS"),
                redis_data_type="hash",
                s3_bucket=config.get("notifications.s3.bucket"),
                s3_key=config.get(
                    "notifications.s3.key", "notifications/all_notifications_v1.json.gz"
                ),
            )
        return {
            "error_count_by_role": error_count,
            "num_new_or_changed_notifications": len(new_or_changed_notifications_l),
        }



"""
Main account stack for running ConsoleMe on ECS
"""

from aws_cdk import core as cdk
from nested_stacks.alb_stack import ALBStack
from nested_stacks.auth_stack import AuthStack
from nested_stacks.cache_stack import CacheStack
from nested_stacks.compute_stack import ComputeStack
from nested_stacks.config_stack import ConfigStack
from nested_stacks.db_stack import DBStack
from nested_stacks.domain_stack import DomainStack
from nested_stacks.iam_stack import IAMStack
from nested_stacks.shared_stack import SharedStack
from nested_stacks.vpc_stack import VPCStack

from consoleme_ecs_cdk.service.constants import BASE_NAME


class ConsolemeEcsServiceStack(cdk.Stack):
    """Main stack for the ConsoleMe service.

    Attributes:
        flight_speed     The maximum speed that such a bird can attain.
        nesting_grounds  The locale where these birds congregate to reproduce.
    """

    def __init__(self, scope: cdk.Construct, id: str, **kwargs) -> None:
        super().__init__(scope, id, **kwargs)

        shared_stack = SharedStack(self, "Shared")

        iam_stack = IAMStack(self, "IAM", s3_bucket=shared_stack.s3_bucket)

        DBStack(self, "DB")

        vpc_stack = VPCStack(self, "VPC")

        alb_stack = ALBStack(
            self, "ALB", vpc=vpc_stack.vpc, consoleme_sg=vpc_stack.consoleme_sg
        )

        domain_stack = DomainStack(
            self, "Domain", consoleme_alb=alb_stack.consoleme_alb
        )

        auth_stack = AuthStack(
            self, "Auth", domain_name=domain_stack.route53_record.domain_name
        )

        cache_stack = CacheStack(
            self, "Cache", vpc=vpc_stack.vpc, redis_sg=vpc_stack.redis_sg
        )

        config_stack = ConfigStack(
            self,
            "Config",
            cognito_user_pool=auth_stack.cognito_user_pool,
            redis=cache_stack.redis,
            domain_name=domain_stack.route53_record.domain_name,
            s3_bucket_name=shared_stack.s3_bucket.bucket_name,
            create_configuration_lambda_role_arn=iam_stack.create_configuration_lambda_role.role_arn,
        )

        compute_stack = ComputeStack(
            self,
            "Compute",
            vpc=vpc_stack.vpc,
            consoleme_alb=alb_stack.consoleme_alb,
            consoleme_sg=vpc_stack.consoleme_sg,
            s3_bucket_name=shared_stack.s3_bucket.bucket_name,
            certificate=domain_stack.certificate,
            task_role_arn=iam_stack.ecs_task_role.role_arn,
            task_execution_role_arn=iam_stack.ecs_task_execution_role.role_arn,
        )

        compute_stack.node.add_dependency(config_stack)

        # Output the service URL to CloudFormation outputs

        cdk.CfnOutput(
            self,
            f"{BASE_NAME}URL",
            value="https://" + domain_stack.route53_record.domain_name,
        )



import base64
import sys
import time
from typing import Any, Callable, Dict, List, Optional, Union

import jsonschema
import jwt
import ujson as json
from cryptography.hazmat.backends.openssl.rsa import _RSAPublicKey

from consoleme.config import config
from consoleme.lib.crypto import Crypto
from consoleme.lib.generic import is_in_group
from consoleme.lib.plugins import get_plugin_by_name

crypto = Crypto()
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
log = config.get_logger()


class AuthenticatedResponse:
    def __init__(self, **kwargs):
        self.authenticated: bool = kwargs.get("authenticated", False)
        self.redirect: str = kwargs.get("redirect", "")

    def get(self, query):
        return self.__dict__.get(query)

    def to_json(self):
        if isinstance(self.value, list):
            self.value = ",".join(self.value)
        d = {
            "name": self.name,
            "attributeType": self.type,
            "attributeValue": self.value,
            "sensitive": self.sensitive,
            "immutable": self.immutable,
        }
        return json.dumps(d)


async def generate_auth_token(
    user, ip, challenge_uuid, expiration=config.get("challenge.token_expiration", 3600)
):
    stats.count("generate_auth_token")
    log_data = {
        "user": user,
        "ip": ip,
        "function": f"{__name__}.{sys._getframe().f_code.co_name}",
        "message": "Generating token for user",
        "challenge_uuid": challenge_uuid,
    }
    log.debug(log_data)
    current_time = int(time.time())
    valid_before = current_time + expiration
    valid_after = current_time

    auth_token = {
        "user": user,
        "ip": ip,
        "challenge_uuid": challenge_uuid,
        "valid_before": valid_before,
        "valid_after": valid_after,
    }

    to_sign = (
        "{{'user': '{0}', 'ip': '{1}', 'challenge_uuid'': '{2}', "
        "'valid_before'': '{3}', 'valid_after'': '{4}'}}"
    ).format(user, ip, challenge_uuid, valid_before, valid_after)

    sig = crypto.sign(to_sign)

    auth_token["sig"] = sig
    return base64.b64encode(json.dumps(auth_token).encode())


async def validate_auth_token(user, ip, token):
    stats.count("validate_auth_token")
    log_data = {
        "user": user,
        "ip": ip,
        "function": f"{__name__}.{sys._getframe().f_code.co_name}",
        "message": "Validating token for user",
    }
    log.debug(log_data)
    if not token:
        stats.count("validate_auth_token.no_token")
        msg = f"No token passed. User: {user}. IP: {ip}."
        log.error(msg, exc_info=True)
        raise Exception(msg)
    decoded_token = base64.b64decode(token)
    auth_token = json.loads(decoded_token)
    current_time = int(time.time())

    if auth_token.get("user") != user:
        stats.count("validate_auth_token.user_mismatch")
        msg = f"Auth token has a different user: {auth_token.get('user')}. User passed to function: {user}"
        log.error(msg, exc_info=True)
        raise Exception(msg)

    if auth_token.get("ip") != ip:
        stats.count("validate_auth_token.ip_mismatch")
        msg = f"Auth token has a different IP: {auth_token.get('ip')}. IP passed to function: {ip}"
        log.error(msg, exc_info=True)
        raise Exception(msg)

    if (
        auth_token.get("valid_before") < current_time
        or auth_token.get("valid_after") > current_time
    ):
        stats.count("validate_auth_token.expiration_error")
        msg = (
            f"Auth token has expired. valid_before: {auth_token.get('valid_before')}. "
            f"valid_after: {auth_token.get('valid_after')}. Current_time: {current_time}"
        )
        log.error(msg, exc_info=True)
        raise Exception(msg)

    to_verify = (
        "{{'user': '{0}', 'ip': '{1}', 'challenge_uuid'': '{2}', "
        "'valid_before'': '{3}', 'valid_after'': '{4}'}}"
    ).format(
        auth_token.get("user"),
        auth_token.get("ip"),
        auth_token.get("challenge_uuid"),
        auth_token.get("valid_before"),
        auth_token.get("valid_after"),
    )

    token_details = {
        "valid": crypto.verify(to_verify, auth_token.get("sig")),
        "user": auth_token.get("user"),
    }

    return token_details


def can_admin_all(user: str, user_groups: List[str]):
    application_admin = config.get("application_admin")
    if application_admin:
        if user == application_admin or application_admin in user_groups:
            return True
    if is_in_group(
        user,
        user_groups,
        [
            *config.get("groups.can_admin", []),
            *config.get("dynamic_config.groups.can_admin", []),
        ],
    ):
        return True
    return False


def can_create_roles(
    user: str,
    user_groups: List[str],
) -> bool:
    if can_admin_all(user, user_groups):
        return True
    if is_in_group(
        user,
        user_groups,
        [
            *config.get("groups.can_create_roles", []),
            *config.get("dynamic_config.groups.can_create_roles", []),
        ],
    ):
        return True
    return False


def can_admin_policies(user: str, user_groups: List[str]) -> bool:
    if can_admin_all(user, user_groups):
        return True
    if is_in_group(
        user,
        user_groups,
        [
            *config.get("groups.can_admin_policies", []),
            *config.get("dynamic_config.groups.can_admin_policies", []),
        ],
    ):
        return True
    return False


def can_delete_iam_principals_app(app_name):
    if app_name in [
        *config.get("groups.can_delete_roles_apps", []),
        *config.get("dynamic_config.groups.can_delete_roles_apps", []),
    ]:
        return True
    return False


def can_delete_iam_principals(
    user: str,
    user_groups: List[str],
) -> bool:
    if can_admin_all(user, user_groups):
        return True
    if is_in_group(
        user,
        user_groups,
        [
            # TODO: Officially deprecate groups.can_delete_roles config key
            *config.get("groups.can_delete_roles", []),
            # TODO: Officially deprecate dynamic_config.groups.can_delete_roles config key
            *config.get("dynamic_config.groups.can_delete_roles", []),
            *config.get("groups.can_delete_iam_principals", []),
            *config.get("dynamic_config.groups.can_delete_iam_principals", []),
        ],
    ):
        return True
    return False


def can_edit_dynamic_config(
    user: str,
    user_groups: List[str],
) -> bool:
    if can_admin_all(user, user_groups):
        return True
    if is_in_group(
        user,
        user_groups,
        [
            *config.get("groups.can_edit_config", []),
            *config.get("dynamic_config.groups.can_edit_config", []),
        ],
    ):
        return True
    return False


def can_edit_attributes(
    user: str, user_groups: List[str], group_info: Optional[Any]
) -> bool:
    if can_admin_all(user, user_groups):
        return True

    if is_in_group(
        user,
        user_groups,
        [
            *config.get("groups.can_admin_restricted", []),
            *config.get("dynamic_config.groups.can_admin_restricted", []),
        ],
    ):
        return True
    if is_in_group(
        user,
        user_groups,
        [
            *config.get("groups.can_edit_attributes", []),
            *config.get("dynamic_config.groups.can_edit_attributes", []),
        ],
    ):
        return True
    return False


def can_modify_members(
    user: str, user_groups: List[str], group_info: Optional[Any]
) -> bool:
    # No users can modify members on restricted groups
    if group_info and group_info.restricted:
        return False

    if can_admin_all(user, user_groups):
        return True

    if is_in_group(
        user,
        user_groups,
        [
            *config.get("groups.can_admin_restricted", []),
            *config.get("dynamic_config.groups.can_admin_restricted", []),
        ],
    ):
        return True

    if is_in_group(
        user,
        user_groups,
        [
            *config.get("groups.can_modify_members", []),
            *config.get("dynamic_config.groups.can_modify_members", []),
        ],
    ):
        return True
    return False


def can_edit_sensitive_attributes(
    user: str, user_groups: List[str], group_info: Optional[Any]
) -> bool:
    if can_admin_all(user, user_groups):
        return True
    if is_in_group(
        user,
        user_groups,
        [
            *config.get("groups.can_edit_sensitive_attributes", []),
            *config.get("dynamic_config.groups.can_edit_sensitive_attributes", []),
        ],
    ):
        return True
    return False


def is_sensitive_attr(attribute):
    for attr in [
        *config.get("groups.attributes.boolean", []),
        *config.get("dynamic_config.groups.attributes.boolean", []),
    ]:
        if attr.get("name") == attribute:
            return attr.get("sensitive", False)

    for attr in [
        *config.get("groups.attributes.list", []),
        *config.get("dynamic_config.groups.attributes.list", []),
    ]:
        if attr.get("name") == attribute:
            return attr.get("sensitive", False)
    return False


class Error(Exception):
    """Base class for exceptions in this module."""


class AuthenticationError(Error):
    """Exception raised for AuthN errors."""

    def __init__(self, message):
        self.message = message


def mk_jwt_validator(
    verification_str: _RSAPublicKey,
    header_cfg: Dict[str, Dict[str, List[str]]],
    payload_cfg: Dict[str, Dict[str, List[str]]],
) -> Callable:
    def validate_jwt(jwt_str):
        try:
            tkn = jwt.decode(
                jwt_str, verification_str, algorithms=header_cfg["alg"]["enum"]
            )
        except jwt.InvalidSignatureError:
            raise AuthenticationError("Invalid Token Signature")
        except jwt.ExpiredSignatureError:
            raise AuthenticationError("Token Expired")
        except jwt.InvalidAudienceError:
            raise AuthenticationError("Invalid Token Audience")
        except jwt.DecodeError:
            raise AuthenticationError("Invalid Token")
        except jwt.InvalidTokenError:
            raise AuthenticationError("Malformed Token")

        try:
            jsonschema.validate(tkn, payload_cfg)
        except jsonschema.ValidationError as e:
            raise AuthenticationError(e.message)
        return tkn

    return validate_jwt


class UnsupportedKeyTypeError(Error):
    """Exception raised unsupported JWK Errors."""

    def __init__(self, message):
        self.message = message


def mk_jwks_validator(
    jwk_set: List[Dict[str, Union[str, List[str]]]],
    header_cfg: Dict[str, Dict[str, List[str]]],
    payload_cfg: Dict[str, Dict[str, List[str]]],
) -> Callable:
    keys = []
    for j in jwk_set:
        if j["kty"] == "RSA":
            j_str = json.dumps(j)
            keys.append(jwt.algorithms.RSAAlgorithm.from_jwk(j_str))
        else:
            raise UnsupportedKeyTypeError("Unsupported Key Type: %s" % j["kty"])

    validators = [mk_jwt_validator(k, header_cfg, payload_cfg) for k in keys]

    def validate_jwt(jwt_str):
        result = None
        for v in validators:
            try:
                result = v(jwt_str)
            except AuthenticationError as e:
                result = e
            else:
                break
        if isinstance(result, Exception):
            raise result
        return result

    return validate_jwt



import re
from typing import Dict, List, Optional

import ujson as json
from policyuniverse.expander_minimizer import _expand_wildcard_action

from consoleme.config import config
from consoleme.exceptions.exceptions import InvalidRequestParameter, MustBeFte
from consoleme.handlers.base import BaseAPIV1Handler, BaseHandler, BaseMtlsHandler
from consoleme.lib.account_indexers import get_account_id_to_name_mapping
from consoleme.lib.cache import retrieve_json_data_from_redis_or_s3
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.lib.redis import redis_get, redis_hgetall

log = config.get_logger()
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
aws = get_plugin_by_name(config.get("plugins.aws", "default_aws"))()
auth = get_plugin_by_name(config.get("plugins.auth", "default_auth"))()


class AutocompleteHandler(BaseAPIV1Handler):
    async def get(self):
        """
        /api/v1/policyuniverse/autocomplete/?prefix=
        ---
        get:
            description: Supplies autocompleted permissions for the ace code editor.
            responses:
                200:
                    description: Returns a list of the matching permissions.
        """

        if config.get("policy_editor.disallow_contractors", True) and self.contractor:
            if self.user not in config.get(
                "groups.can_bypass_contractor_restrictions", []
            ):
                raise MustBeFte("Only FTEs are authorized to view this page.")

        only_filter_services = False

        if (
            self.request.arguments.get("only_filter_services")
            and self.request.arguments.get("only_filter_services")[0].decode("utf-8")
            == "true"
        ):
            only_filter_services = True

        prefix = self.request.arguments.get("prefix")[0].decode("utf-8") + "*"
        results = _expand_wildcard_action(prefix)
        if only_filter_services:
            # We return known matching services in a format that the frontend expects to see them. We omit the wildcard
            # character returned by policyuniverse.
            services = sorted(list({r.split(":")[0].replace("*", "") for r in results}))
            results = [{"title": service} for service in services]
        else:
            results = [dict(permission=r) for r in results]
        self.write(json.dumps(results))
        await self.finish()


async def filter_resources(filter, resources, max=20):
    if filter:
        regexp = re.compile(r"{}".format(filter.strip()), re.IGNORECASE)
        results: List[str] = []
        for resource in resources:
            try:
                if regexp.search(str(resource.get(filter))):
                    if len(results) == max:
                        return results
                    results.append(resource)
            except re.error:
                # Regex error. Return no results
                pass
        return results
    else:
        return resources


async def handle_resource_type_ahead_request(cls):
    try:
        search_string: str = cls.request.arguments.get("search")[0].decode("utf-8")
    except TypeError:
        cls.send_error(400, message="`search` parameter must be defined")
        return

    try:
        resource_type: str = cls.request.arguments.get("resource")[0].decode("utf-8")
    except TypeError:
        cls.send_error(400, message="`resource_type` parameter must be defined")
        return

    account_id = None
    topic_is_hash = True
    account_id_optional: Optional[List[bytes]] = cls.request.arguments.get("account_id")
    if account_id_optional:
        account_id = account_id_optional[0].decode("utf-8")

    limit: int = 10
    limit_optional: Optional[List[bytes]] = cls.request.arguments.get("limit")
    if limit_optional:
        limit = int(limit_optional[0].decode("utf-8"))

    # By default, we only return the S3 bucket name of a resource and not the full ARN
    # unless you specifically request it
    show_full_arn_for_s3_buckets: Optional[bool] = cls.request.arguments.get(
        "show_full_arn_for_s3_buckets"
    )

    role_name = False
    if resource_type == "s3":
        topic = config.get("redis.s3_bucket_key", "S3_BUCKETS")
        s3_bucket = config.get("account_resource_cache.s3_combined.bucket")
        s3_key = config.get(
            "account_resource_cache.s3_combined.file",
            "account_resource_cache/cache_s3_combined_v1.json.gz",
        )
    elif resource_type == "sqs":
        topic = config.get("redis.sqs_queues_key", "SQS_QUEUES")
        s3_bucket = config.get("account_resource_cache.sqs_combined.bucket")
        s3_key = config.get(
            "account_resource_cache.sqs_combined.file",
            "account_resource_cache/cache_sqs_queues_combined_v1.json.gz",
        )
    elif resource_type == "sns":
        topic = config.get("redis.sns_topics_key ", "SNS_TOPICS")
        s3_bucket = config.get("account_resource_cache.sns_topics_combined.bucket")
        s3_key = config.get(
            "account_resource_cache.sns_topics_topics_combined.file",
            "account_resource_cache/cache_sns_topics_combined_v1.json.gz",
        )
    elif resource_type == "iam_arn":
        topic = config.get("aws.iamroles_redis_key ", "IAM_ROLE_CACHE")
        s3_bucket = config.get(
            "cache_iam_resources_across_accounts.all_roles_combined.s3.bucket"
        )
        s3_key = config.get(
            "cache_iam_resources_across_accounts.all_roles_combined.s3.file",
            "account_resource_cache/cache_all_roles_v1.json.gz",
        )
    elif resource_type == "iam_role":
        topic = config.get("aws.iamroles_redis_key ", "IAM_ROLE_CACHE")
        s3_bucket = config.get(
            "cache_iam_resources_across_accounts.all_roles_combined.s3.bucket"
        )
        s3_key = config.get(
            "cache_iam_resources_across_accounts.all_roles_combined.s3.file",
            "account_resource_cache/cache_all_roles_v1.json.gz",
        )
        role_name = True
    elif resource_type == "account":
        topic = None
        s3_bucket = None
        s3_key = None
        topic_is_hash = False
    elif resource_type == "app":
        topic = config.get("celery.apps_to_roles.redis_key", "APPS_TO_ROLES")
        s3_bucket = None
        s3_key = None
        topic_is_hash = False
    else:
        cls.send_error(404, message=f"Invalid resource_type: {resource_type}")
        return

    if not topic and resource_type != "account":
        raise InvalidRequestParameter("Invalid resource_type specified")

    if topic and topic_is_hash and s3_key:
        data = await retrieve_json_data_from_redis_or_s3(
            redis_key=topic, redis_data_type="hash", s3_bucket=s3_bucket, s3_key=s3_key
        )
    elif topic:
        data = await redis_get(topic)

    results: List[Dict] = []

    unique_roles: List[str] = []

    if resource_type == "account":
        account_and_id_list = []
        account_ids_to_names = await get_account_id_to_name_mapping()
        for account_id, account_name in account_ids_to_names.items():
            account_and_id_list.append(f"{account_name} ({account_id})")
        for account in account_and_id_list:
            if search_string.lower() in account.lower():
                results.append({"title": account})
    elif resource_type == "app":
        results = {}
        all_role_arns = []
        all_role_arns_j = await redis_hgetall(
            (config.get("aws.iamroles_redis_key", "IAM_ROLE_CACHE"))
        )
        if all_role_arns_j:
            all_role_arns = all_role_arns_j.keys()
        # ConsoleMe (Account: Test, Arn: arn)
        # TODO: Make this OSS compatible and configurable
        try:
            accounts = await get_account_id_to_name_mapping()
        except Exception as e:  # noqa
            accounts = {}

        app_to_role_map = {}
        if data:
            app_to_role_map = json.loads(data)
        seen: Dict = {}
        seen_roles = {}
        for app_name, roles in app_to_role_map.items():
            if len(results.keys()) > 9:
                break
            if search_string.lower() in app_name.lower():
                results[app_name] = {"name": app_name, "results": []}
                for role in roles:
                    account_id = role.split(":")[4]
                    account = accounts.get(account_id, "")
                    parsed_app_name = (
                        f"{app_name} on {account} ({account_id}) ({role})]"
                    )
                    if seen.get(parsed_app_name):
                        continue
                    seen[parsed_app_name] = True
                    seen_roles[role] = True
                    results[app_name]["results"].append(
                        {"title": role, "description": account}
                    )
        for role in all_role_arns:
            if len(results.keys()) > 9:
                break
            if search_string.lower() in role.lower():
                if seen_roles.get(role):
                    continue
                account_id = role.split(":")[4]
                account = accounts.get(account_id, "")
                if not results.get("Unknown App"):
                    results["Unknown App"] = {"name": "Unknown App", "results": []}
                results["Unknown App"]["results"].append(
                    {"title": role, "description": account}
                )

    else:
        if not data:
            return []
        for k, v in data.items():
            if account_id and k != account_id:
                continue
            if role_name:
                try:
                    r = k.split("role/")[1]
                except IndexError:
                    continue
                if search_string.lower() in r.lower():
                    if r not in unique_roles:
                        unique_roles.append(r)
                        results.append({"title": r})
            elif resource_type == "iam_arn":
                if k.startswith("arn:") and search_string.lower() in k.lower():
                    results.append({"title": k})
            else:
                list_of_items = json.loads(v)
                for item in list_of_items:
                    # A Hack to get S3 to show full ARN, and maintain backwards compatibility
                    # TODO: Fix this in V2 of resource specific typeahead endpoints
                    if resource_type == "s3" and show_full_arn_for_s3_buckets:
                        item = f"arn:aws:s3:::{item}"
                    if search_string.lower() in item.lower():
                        results.append({"title": item, "account_id": k})
                    if len(results) > limit:
                        break
            if len(results) > limit:
                break
    return results


class ApiResourceTypeAheadHandler(BaseMtlsHandler):
    async def get(self):
        if self.requester["name"] not in config.get("api_auth.valid_entities", []):
            raise Exception("Call does not originate from a valid API caller")
        results = await handle_resource_type_ahead_request(self)
        self.write(json.dumps(results))


class ResourceTypeAheadHandler(BaseHandler):
    async def get(self):
        if config.get("policy_editor.disallow_contractors", True) and self.contractor:
            if self.user not in config.get(
                "groups.can_bypass_contractor_restrictions", []
            ):
                raise MustBeFte("Only FTEs are authorized to view this page.")
        results = await handle_resource_type_ahead_request(self)
        self.write(json.dumps(results))



from typing import Dict, Optional, Union

from consoleme.default_plugins.plugins.metrics.base_metric import Metric


class DefaultMetric(Metric):
    def count(self, metric_name, tags=None):
        # TODO(ccastrapel): Having Cloudwatch, Graphite, or other configurtable default metric sources here would be a
        # good idea.
        pass

    def gauge(self, metric_name, metric_value, tags=None):
        pass

    def timer(
        self,
        metric_name: str,
        tags: Optional[Union[Dict[str, Union[str, bool]], Dict[str, str]]] = None,
    ) -> None:
        pass



from typing import Dict, List

from consoleme.config import config


async def get_custom_page_header(user: str, user_groups: List[str]) -> Dict[str, str]:
    """
    Args:
        user: The user's e-mail address
        user_groups: the user's group memberships

    Returns:
        Headers to show on the page. These headers can be specific to the user's group memberships, or the user's
        email address, or generic. If a configuration is not set, a header will not be shown.

    Example:
        Visit https://YOUR_CONSOLEME_DOMAIN/config
        Add this to the page:
        ```
        custom_headers_for_group_members:
          - users_or_groups:
             - you@example.com
             - a_group@example.com
            title: Important message!
            message: Read this!
        ```
    """
    if config.get("example_config.is_example_config", False):
        return {
            "custom_header_message_title": config.get("example_config.title"),
            "custom_header_message_text": config.get("example_config.text"),
            "custom_header_message_route": config.get("example_config.routes"),
        }
    custom_headers_for_group_members = config.get(
        "dynamic_config.custom_headers_for_group_members", []
    )
    for custom_header in custom_headers_for_group_members:
        for header_group in custom_header.get("users_or_groups", []):
            if header_group in user_groups or user == header_group:
                return {
                    "custom_header_message_title": custom_header.get("title", ""),
                    "custom_header_message_text": custom_header.get("message", ""),
                    "custom_header_message_route": custom_header.get("route", ".*"),
                }
    return {
        "custom_header_message_title": config.get(
            "headers.custom_header_message.title", ""
        ),
        "custom_header_message_text": config.get(
            "headers.custom_header_message.text", ""
        ),
        "custom_header_message_route": config.get(
            "headers.custom_header_message.route", ".*"
        ),
    }



"""
Cache stack for running ConsoleMe on ECS
"""

from aws_cdk import aws_ec2 as ec2
from aws_cdk import aws_elasticache as ec
from aws_cdk import core as cdk


class CacheStack(cdk.NestedStack):
    """
    Cache stack for running ConsoleMe on ECS
    """

    def __init__(
        self,
        scope: cdk.Construct,
        id: str,
        vpc: ec2.Vpc,
        redis_sg: ec2.SecurityGroup,
        **kwargs
    ) -> None:
        super().__init__(scope, id, **kwargs)

        # Redis node

        subnet_ids = []
        for subnet in vpc.private_subnets:
            subnet_ids.append(subnet.subnet_id)

        redis_subnet_group = ec.CfnSubnetGroup(
            self,
            "RedisSubnetGroup",
            cache_subnet_group_name="redis-subnet-group",
            description="Subnet group for Redis Cluster",
            subnet_ids=subnet_ids,
        )

        redis = ec.CfnCacheCluster(
            self,
            "RedisCluster",
            cache_node_type="cache.t3.micro",
            engine="redis",
            engine_version="6.x",
            num_cache_nodes=1,
            auto_minor_version_upgrade=True,
            cache_subnet_group_name=redis_subnet_group.ref,
            vpc_security_group_ids=[redis_sg.security_group_id],
        )

        self.redis = redis



from enum import Enum
from typing import Any, Dict, List, Optional, Set

from pydantic import Field

from consoleme.lib.pydantic import BaseModel


class ConsoleMeUserNotificationAction(BaseModel):
    http_method: str = Field(
        ..., description="HTTP method for the request", example="get"
    )
    uri: str = Field(
        ..., description="URL to send the request if user invokes the action"
    )
    text: str = Field(..., description="Text to show for the call-to-action")


class ConsoleMeUserNotification(BaseModel):
    predictable_id: str = Field(
        ...,
        description=(
            "Predictable ID for this notification. Generally used to prevent duplicate notifications"
        ),
    )
    type: str = Field(
        ...,
        description=(
            "Each class of notification should be given a separate type, "
            "IE: `cloudtrail_generated_policy` or `proxy_generated_policy"
        ),
    )
    users_or_groups: Set[str] = Field(
        ..., description="Users or groups who should see the notification"
    )
    event_time: int = Field(..., description="Time that the event took place")
    expiration: int = Field(
        ..., description="Time that this entry should stop notifying people"
    )
    expired: bool = Field(
        ...,
        description="A more obvious indicator about whether a notification has expired",
    )
    header: Optional[str] = Field(
        None, description="Bolded text (Header) for the notification"
    )
    message: str = Field(
        ...,
        description="An (optionally markdown) formatted message to show to users in the UI",
    )
    message_actions: List[ConsoleMeUserNotificationAction] = Field(
        ...,
        description=(
            "An optional list of actions to make available to the user. These will be shown below the notification."
        ),
    )
    details: Dict[str, Any] = Field(
        ...,
        description=(
            "Extra details about the notification for power users. "
            "This will be accessible to the user in the UI, but not visible by default"
        ),
    )
    read_by_users: List[str] = Field(
        ...,
        description=(
            "List of users the notification is `read` for. It will show up in the user's list of notifications, "
            "but will not be shown as `unread` nor be included in the unread counter."
        ),
    )
    read_by_all: bool = Field(
        False,
        description=(
            "Notification is `marked as read` for all users and will not appear as an unread notification for "
            "any users."
        ),
    )
    hidden_for_users: List[str] = Field(
        ...,
        description=(
            "List of users the notification is `hidden` for. It will not appear at all in the user's list of "
            "notifications."
        ),
    )
    hidden_for_all: bool = Field(
        False,
        description=(
            "Notification is `marked as hidden` for all users, and will not appear in the notificaiton list for "
            "any user."
        ),
    )
    read_for_current_user: bool = Field(
        None,
        description=(
            "Convenience variable  to mark a notification as read for current user when retrieving notification. "
            "This makes it so the frontend doesn't need to do any computation to figure it out. A Non-None value "
            "should not be stored in the database."
        ),
    )
    version: int = Field(..., description=("Version of the notification model"))


class ConsoleMeNotificationUpdateAction(Enum):
    toggle_read_for_current_user = "toggle_read_for_current_user"
    toggle_read_for_all_users = "toggle_read_for_all_users"
    toggle_hidden_for_current_user = "toggle_hidden_for_current_user"
    toggle_hidden_for_all_users = "toggle_hidden_for_all_users"


class ConsoleMeNotificationUpdateRequest(BaseModel):
    action: ConsoleMeNotificationUpdateAction
    notifications: List[ConsoleMeUserNotification]


class GetNotificationsForUserResponse(BaseModel):
    notifications: List[ConsoleMeUserNotification]
    unread_count: int



import pkgutil
from typing import Any

import pkg_resources

global_plugins = {}


def iter_namespace(ns_pkg):
    # Specifying the second argument (prefix) to iter_modules makes the
    # returned name an absolute name instead of a relative one. This allows
    # import_module to work without having to do additional modification to
    # the name.
    return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")


def get_plugin_by_name(plugin_name: str) -> Any:
    if global_plugins.get(plugin_name):
        return global_plugins[plugin_name]
    plugins = []
    for ep in pkg_resources.iter_entry_points("consoleme.plugins"):
        plugins.append(ep.name)
        if ep.name == plugin_name:
            global_plugins[ep.name] = ep.load()
            return global_plugins[ep.name]
    initial_exception_message = f"Could not find the specified plugin: {plugin_name}. "
    if plugin_name == "default_config":
        initial_exception_message = (
            f"Could not find the specified plugin: {plugin_name}. "
            "Please install it with `pip install -e default_plugins` "
            "from the ConsoleMe directory. "
        )

    exception_message = (
        initial_exception_message + f"Plugins found: {', '.join(plugins)}. "
        f"Make sure you've set the environment variable CONSOLEME_CONFIG_ENTRYPOINT to the name of your configuration "
        f"entrypoint, otherwise it will default to `default_config`."
    )

    raise Exception(exception_message)


def import_class_by_name(class_full_path: str):
    """
    Import a class by a dot-delimited class name.
    i.e: import_class("consoleme.default_plugins.plugins.metrics.default_metrics.DefaultMetric")
        --> <class 'consoleme.default_plugins.plugins.metrics.default_metrics.DefaultMetric'>
    """

    d = class_full_path.rfind(".")
    class_name = class_full_path[d + 1 : len(class_full_path)]
    m = __import__(class_full_path[0:d], globals(), locals(), [class_name])
    return getattr(m, class_name)



"""
Shared stack for running ConsoleMe on ECS
"""

from aws_cdk import aws_s3 as s3
from aws_cdk import core as cdk


class SharedStack(cdk.NestedStack):
    """
    Shared stack for running ConsoleMe on ECS
    """

    def __init__(self, scope: cdk.Construct, id: str, **kwargs) -> None:
        super().__init__(scope, id, **kwargs)

        s3_bucket = s3.Bucket(
            self,
            "ConfigBucket",
            removal_policy=cdk.RemovalPolicy.DESTROY,
            block_public_access=s3.BlockPublicAccess.BLOCK_ALL,
            encryption=s3.BucketEncryption.S3_MANAGED,
        )

        self.s3_bucket = s3_bucket



from consoleme.default_plugins.plugins.internal_routes.handlers.internal_demo_route import (
    InternalDemoRouteHandler,
)


class InternalRoutes:
    ui_modules = {}

    def get_internal_routes(self, make_jwt_validator, jwt_validator=None):
        # The below code can be used with your ConsoleMe Internal package name to generate a path to your internal
        # JavaScript and HTML files, if you wish to render these for the handler.
        # path = pkg_resources.resource_filename("consoleme_internal", "templates")
        internal_routes = [
            (r"/internal_demo_route/?", InternalDemoRouteHandler),
            # An example of serving static internal content is below, which would make use of the template path variable
            # You defined above.
            # (
            #     r"/static_internal/(.*)",
            #     NoCacheStaticFileHandler,
            #     dict(path=os.path.join(path, "static")),
            # ),
        ]
        return internal_routes


def init():
    """Initialize the internal routes plugin."""
    return InternalRoutes()



"""
example_consoleme_plugins
==============

This is an example setup configuration for ConsoleMe plugins. You can get
started writing your own plugins by copying this directory (default_plugins)
and customizing it for your needs. Be sure to update the "default_*" entrypoint
names below.

Overview of steps:
1. Copy default_plugins directory to your desired location
2. Change to new default_plugins directory
3. Delete __init__.py
4. Modify plugins to your liking
5. Update the entrypoints in setup.py, e.g. default_config -> my_custom_config
6. Update your configuration to use the new plugin entrypoints
"""
from setuptools import find_packages, setup

setup(
    name="example_consoleme_plugins",
    version="0.0.1",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    setup_requires=["setupmeta"],
    python_requires=">=3.8",
    entry_points={
        "consoleme.plugins": [
            # Change the name of the entry point for your plugin here, and in your configuration
            "default_config = example_consoleme_plugins.plugins.config.config:Config",
            "default_auth = example_consoleme_plugins.plugins.auth.auth:Auth",
            "default_aws = example_consoleme_plugins.plugins.aws.aws:Aws",
            "default_celery_tasks = example_consoleme_plugins.plugins.celery_tasks.celery_tasks:internal_schedule",
            "default_celery_tasks_functions = example_consoleme_plugins.plugins.celery_tasks.celery_tasks:CeleryTasks",
            "default_metrics = example_consoleme_plugins.plugins.metrics.metrics:Metric",
            "default_policies = example_consoleme_plugins.plugins.policies.policies:Policies",
            "default_group_mapping = example_consoleme_plugins.plugins.group_mapping.group_mapping:GroupMapping",
            "default_internal_routes = example_consoleme_plugins.plugins.internal_routes.internal_routes:InternalRoutes",
        ]
    },
)



import random
import re
import string
from datetime import datetime
from random import randint
from typing import Any, Dict, List, Optional, Union
from urllib.parse import unquote_plus

import pandas as pd
import ujson as json
from dateutil import parser

from consoleme.config import config
from consoleme.exceptions.exceptions import MissingRequestParameter
from consoleme.models import (
    AwsResourcePrincipalModel,
    HoneybeeAwsResourceTemplatePrincipalModel,
)


def str2bool(v: Optional[Union[bool, str]]) -> bool:
    if isinstance(v, bytes):
        v = v.decode()
    if not v:
        return False
    if type(v) is bool and v is True:
        return True
    return v.lower() in ["true", "True"]


# Yield successive n-sized
# chunks from list list_.
def divide_chunks(list_, n):
    """
    Yields successive n=zied chunks from list l by looping
    until length l.

    `divide_chunks(["a","b","c","d","e"], 2)` yields:
    ['a', 'b', 'c']
    ['d', 'e']
    """
    for i in range(0, len(list_), n):
        yield list_[i : i + n]


def generate_html(d: List[Dict[str, Union[str, bool]]]) -> str:
    """
    Pass in a dict with a list of rows to include in a formatted table. This will return the HTML for the table.

    :param d:
    :return:
        html: HTML formatted table
    """
    if not d:
        return
    pd.set_option("display.max_colwidth", -1)
    df = pd.DataFrame(d)
    html = df.to_html(classes=["ui", "celled", "table"], escape=False, index=False)
    return html


def auto_split(s: str) -> List[str]:
    results = []

    for i in s.splitlines():
        results.extend(i.split(","))
    return results


def is_valid_role_arn(arn: str) -> bool:
    # This is valid enough as far as we are concerned.
    if not arn.startswith("arn:aws:iam::"):
        return False
    return True


def regex_filter(
    filter: Dict[str, str], items: List[Dict[str, Union[str, None, bool]]]
) -> List[Dict[str, Union[str, None, bool]]]:
    if filter.get("filter"):
        results = []
        if filter.get("type", "") == "date":
            from_date = None
            to_date = None
            try:
                if filter.get("from_date"):
                    from_date = parser.parse(filter.get("from_date"))
                if filter.get("to_date"):
                    to_date = parser.parse(filter.get("to_date"))
                if not from_date and not to_date:
                    return items
            except:  # noqa
                # Unable to parse date. Return items.
                return results
            for item in items:
                item_date = parser.parse(
                    item.get(filter.get("field"))
                )  # What if invalid date
                if from_date and to_date and from_date <= item_date <= to_date:
                    results.append(item)
                    continue
                if from_date and not to_date and item_date >= from_date:
                    results.append(item)
                    continue
                if to_date and not from_date and item_date <= to_date:
                    results.append(item)
                    continue
            return results
        else:
            regexp = re.compile(r"{}".format(filter.get("filter")), re.IGNORECASE)
            for item in items:
                try:
                    if regexp.search(item.get(filter.get("field"))):
                        results.append(item)
                except re.error:
                    # Regex error. Return no results
                    pass
            return results
    else:
        return items


def is_in_group(
    user: str, user_groups: List[str], required_groups: Union[List[str], str]
) -> bool:
    if isinstance(required_groups, str):
        required_groups = [required_groups]
    for group in required_groups:
        if group in user_groups or user == group:
            return True
    return False


async def write_json_error(message, obj):
    result = {"status": "error", "message": message}
    obj.write(json.dumps(result))
    obj.finish()


async def sort_nested_dictionary_lists(d):
    for k, v in d.items():
        if isinstance(v, list):
            for i in range(0, len(v)):
                if isinstance(v[i], dict):
                    v[i] = await sort_nested_dictionary_lists(v[i])
                d[k] = sorted(v)
        if isinstance(v, dict):
            d[k] = await sort_nested_dictionary_lists(v)
    return d


def is_in_time_range(t, time_range):
    valid_days = time_range.get("days")
    if t.weekday() not in valid_days:
        return False
    valid_start_time = t.replace(
        hour=time_range.get("hour_start", 0),
        minute=time_range.get("minute_start", 0),
        second=0,
        microsecond=0,
    )
    valid_end_time = t.replace(
        hour=time_range.get("hour_end", 0),
        minute=time_range.get("minute_end", 0),
        second=0,
        microsecond=0,
    )
    if t < valid_start_time or t > valid_end_time:
        return False
    return True


async def get_random_security_logo():
    if config.get("consoleme_logo.image"):
        return config.get("consoleme_logo.image")
    month = datetime.now().month
    summer = month in [6, 7, 8]

    dir = "sunglasses" if summer else "nosunglasses"
    file = f"{randint(1, 3)}.png"  # nosec
    return f"/images/logos/{dir}/{file}"


async def generate_random_string(string_length=4):
    letters = string.ascii_lowercase
    return "".join(random.choice(letters) for i in range(string_length))  # nosec


async def filter_table(filter_key, filter_value, data):
    if not (filter_key and filter_value):
        # Filter parameters are incorrect. Don't filter
        return data
    results = []
    if isinstance(filter_value, str):
        try:
            regexp = re.compile(r"{}".format(str(filter_value).strip()), re.IGNORECASE)
        except:  # noqa
            # Regex is incorrect. Don't filter
            return data

        for d in data:
            try:
                if regexp.search(str(d.get(filter_key))):
                    results.append(d)
            except re.error:
                # Regex error. Return no results
                pass
        return results
    elif (
        isinstance(filter_value, list)
        and len(filter_value) == 2
        and isinstance(filter_value[0], int)
        and isinstance(filter_value[1], int)
    ):
        # Handles epoch time filter. We expect a start_time and an end_time in
        # a list of elements, and they should be integers
        for d in data:
            if filter_value[0] < int(d.get(filter_key)) < filter_value[1]:
                results.append(d)
        return results


async def iterate_and_format_dict(d: Dict, replacements: Dict):
    """
    Iterates through the values of a dictionary (with or without nested dictionaries), and formats values accordingly
    if they exist in the `replacements` dictionary.

    Example args:
        d = {"something": {"nested": "1{thing}1"},
        replacements = {"thing": "toreplace", "thing2": "dontreplace"}
    Returns: {"something": {"nested": "1toreplace1"}


    :param d:
    :param replacements:
    :return:
    """
    for k, v in d.items():
        if isinstance(v, dict):
            await iterate_and_format_dict(v, replacements)
        else:
            try:
                d[k] = v.format(**replacements)
            except KeyError:
                pass
    return d


async def should_force_redirect(req):
    """
    ConsoleMe should only force a 302 redirect for non-XHR requests
    """
    if req.headers.get("X-Requested-With", "") == "XMLHttpRequest":
        return False
    if req.headers.get("Accept") == "application/json":
        return False
    return True


def sort_dict(original):
    """Recursively sorts dictionary keys and dictionary values in alphabetical order"""
    if isinstance(original, dict):
        res = (
            dict()
        )  # Make a new "ordered" dictionary. No need for Collections in Python 3.7+
        for k, v in sorted(original.items()):
            res[k] = v
        d = res
    else:
        d = original
    for k in d:
        if isinstance(d[k], str):
            continue
        if isinstance(d[k], list) and len(d[k]) > 1 and isinstance(d[k][0], str):
            d[k] = sorted(d[k])
        if isinstance(d[k], dict):
            d[k] = sort_dict(d[k])
        if isinstance(d[k], list) and len(d[k]) >= 1 and isinstance(d[k][0], dict):
            for i in range(len(d[k])):
                d[k][i] = sort_dict(d[k][i])
    return d


def un_wrap_json(json_obj: Any) -> Any:
    """Helper function to unwrap nested JSON in the AWS Config resource configuration."""
    # pylint: disable=C0103,W0703,R0911
    # Is this a field that we can safely return?
    if isinstance(json_obj, (type(None), int, bool, float)):  # noqa
        return json_obj
    # Is this a Datetime? Convert it to a string and return it:
    if isinstance(json_obj, datetime):
        return str(json_obj)
    # Is this a Dictionary?
    if isinstance(json_obj, dict):
        decoded = {}
        for k, v in json_obj.items():
            decoded[k] = un_wrap_json(v)
    # Is this a List?
    elif isinstance(json_obj, list):
        decoded = []
        for x in json_obj:
            decoded.append(un_wrap_json(x))
        # Yes, try to sort the contents of lists. This is because AWS does not consistently store list ordering for many resource types:
        try:
            sorted_list = sorted(decoded)
            decoded = sorted_list
        except Exception:  # noqa  # nosec   # If we can't sort then NBD
            pass
    else:
        # Try to load the JSON string:
        try:
            # Check if the string starts with a "[" or a "{" (because apparently '123' is a valid JSON)
            for check_field in {
                "{",
                "[",
                '"{',
                '"[',
            }:  # Some of the double-wrapping is really ridiculous
                if json_obj.startswith(check_field):
                    decoded = json.loads(json_obj)
                    # If we loaded this properly, then we need to pass the decoded JSON back in for all the nested stuff:
                    return un_wrap_json(decoded)
            # Check if this string is URL Encoded - if it is, then re-run it through:
            decoded = unquote_plus(json_obj)
            if decoded != json_obj:
                return un_wrap_json(decoded)
            return json_obj
        # If we didn't get a JSON back (exception), then just return the raw value back:
        except Exception:  # noqa
            return json_obj
    return decoded


def un_wrap_json_and_dump_values(json_obj: Any) -> Any:
    json_obj = un_wrap_json(json_obj)
    for k, v in json_obj.items():
        json_obj[k] = json.dumps(v)
    return json_obj


async def get_principal_friendly_name(principal):
    if isinstance(principal, HoneybeeAwsResourceTemplatePrincipalModel):
        return principal.resource_identifier
    if isinstance(principal, AwsResourcePrincipalModel):
        return principal.principal_arn
    raise MissingRequestParameter("Unable to determine principal")


class Struct:
    def __init__(self, **entries):
        self.__dict__.update(entries)



"""Handle the base."""
import asyncio
import time
import traceback
import uuid
from datetime import datetime, timedelta
from typing import Any, Union

import pytz
import redis
import tornado.httpclient
import tornado.httputil
import tornado.web
import ujson as json
from tornado import httputil

from consoleme.config import config
from consoleme.exceptions.exceptions import (
    InvalidCertificateException,
    MissingCertificateException,
    MissingConfigurationValue,
    NoGroupsException,
    NoUserException,
    SilentException,
    WebAuthNError,
)
from consoleme.lib.alb_auth import authenticate_user_by_alb_auth
from consoleme.lib.auth import AuthenticationError
from consoleme.lib.jwt import generate_jwt_token, validate_and_return_jwt_token
from consoleme.lib.oidc import authenticate_user_by_oidc
from consoleme.lib.plugins import get_plugin_by_name
from consoleme.lib.redis import RedisHandler
from consoleme.lib.saml import authenticate_user_by_saml
from consoleme.lib.tracing import ConsoleMeTracer

log = config.get_logger()
stats = get_plugin_by_name(config.get("plugins.metrics", "default_metrics"))()
auth = get_plugin_by_name(config.get("plugins.auth", "default_auth"))()
group_mapping = get_plugin_by_name(
    config.get("plugins.group_mapping", "default_group_mapping")
)()


class TornadoRequestHandler(tornado.web.RequestHandler):
    def get_request_ip(self):
        trusted_remote_ip_header = config.get("auth.remote_ip.trusted_remote_ip_header")
        if trusted_remote_ip_header:
            return self.request.headers[trusted_remote_ip_header].split(",")[0]
        return self.request.remote_ip


class BaseJSONHandler(TornadoRequestHandler):
    # These methods are returned in OPTIONS requests.
    # Default methods can be overridden by setting this variable in child classes.
    allowed_methods = ["GET", "HEAD", "PUT", "PATCH", "POST", "DELETE"]

    def __init__(self, *args, **kwargs):
        self.jwt_validator = kwargs.pop("jwt_validator", None)
        self.auth_required = kwargs.pop("auth_required", True)
        if self.jwt_validator is None:
            raise TypeError("Missing required keyword arg jwt_validator")
        super().__init__(*args, **kwargs)

    def check_xsrf_cookie(self):
        # CSRF token is not needed since this is protected by raw OIDC tokens
        pass

    def options(self, *args):
        self.set_header(
            "Access-Control-Allow-Headers",
            self.request.headers["Access-Control-Request-Headers"],
        )
        self.set_header("Content-Length", "0")
        self.set_status(204)
        self.finish()

    async def prepare(self):
        stats.timer("base_handler.incoming_request")
        if self.request.method.lower() == "options":
            return
        self.request_uuid = str(uuid.uuid4())
        payload = self.get_current_user()
        self.auth_context = payload
        self.user = payload["email"]

    def set_default_headers(self, *args, **kwargs):
        self.set_header("Access-Control-Allow-Origin", "*")
        self.set_header("Access-Control-Allow-Methods", ",".join(self.allowed_methods))
        self.set_header("Access-Control-Allow-Credentials", "true")
        self.set_header("Content-Type", "application/json")

    def write_error(self, status_code, **kwargs):
        self.set_header("Content-Type", "application/problem+json")
        title = httputil.responses.get(status_code, "Unknown")
        message = kwargs.get("message", self._reason)
        # self.set_status() modifies self._reason, so this call should come after we grab the reason
        self.set_status(status_code)
        self.finish(
            json.dumps(
                {"status": status_code, "title": title, "message": message}
            )  # noqa
        )

    def get_current_user(self):
        try:
            if config.get("development") and config.get("json_authentication_override"):
                return config.get("json_authentication_override")
            tkn_header = self.request.headers["authorization"]
        except KeyError:
            raise WebAuthNError(reason="Missing Authorization Header")
        else:
            tkn_str = tkn_header.split(" ")[-1]
        try:
            tkn = self.jwt_validator(tkn_str)
        except AuthenticationError as e:
            raise WebAuthNError(reason=e.message)
        else:
            return tkn


class BaseHandler(TornadoRequestHandler):
    """Default BaseHandler."""

    def log_exception(self, *args, **kwargs):
        if args[0].__name__ == "SilentException":
            pass
        else:
            super(BaseHandler, self).log_exception(*args, **kwargs)

    def write_error(self, status_code: int, **kwargs: Any) -> None:
        if self.settings.get("serve_traceback") and "exc_info" in kwargs:
            # in debug mode, try to send a traceback
            self.set_header("Content-Type", "text/plain")
            for line in traceback.format_exception(*kwargs["exc_info"]):
                self.write(line)
            self.finish()
        else:
            self.finish(
                "<html><title>%(code)d: %(message)s</title>"
                "<body>%(code)d: %(message)s</body></html>"
                % {
                    "code": status_code,
                    "message": f"{self._reason} - {config.get('errors.custom_website_error_message', '')}",
                }
            )

    def data_received(self, chunk):
        """Receives the data."""
        pass

    def initialize(self, **kwargs) -> None:
        self.kwargs = kwargs
        self.tracer = None
        self.responses = []
        super(BaseHandler, self).initialize()

    async def prepare(self) -> None:
        self.tracer = None
        await self.configure_tracing()

        if config.get("tornado.xsrf", True):
            cookie_kwargs = config.get("tornado.xsrf_cookie_kwargs", {})
            self.set_cookie(
                config.get("xsrf_cookie_name", "_xsrf"),
                self.xsrf_token,
                **cookie_kwargs,
            )
        self.request_uuid = str(uuid.uuid4())
        stats.timer("base_handler.incoming_request")
        return await self.authorization_flow()

    def write(self, chunk: Union[str, bytes, dict]) -> None:
        if config.get("_security_risk_full_debugging.enabled"):
            if not hasattr(self, "responses"):
                self.responses = []
            self.responses.append(chunk)
        super(BaseHandler, self).write(chunk)

    async def configure_tracing(self):
        self.tracer = ConsoleMeTracer()
        primary_span_name = "{0} {1}".format(
            self.request.method.upper(), self.request.path
        )
        tracer_tags = {
            "http.host": config.hostname,
            "http.method": self.request.method.upper(),
            "http.path": self.request.path,
            "ca": self.get_request_ip(),  # Client IP
            "http.url": self.request.full_url(),
        }
        tracer = await self.tracer.configure_tracing(
            primary_span_name, tags=tracer_tags
        )
        if tracer:
            for k, v in tracer.headers.items():
                self.set_header(k, v)

    def on_finish(self) -> None:
        if hasattr(self, "tracer") and self.tracer:
            asyncio.ensure_future(
                self.tracer.set_additional_tags({"http.status_code": self.get_status()})
            )
            asyncio.ensure_future(self.tracer.finish_spans())
            asyncio.ensure_future(self.tracer.disable_tracing())

        if config.get("_security_risk_full_debugging.enabled"):
            responses = None
            if hasattr(self, "responses"):
                responses = self.responses
            request_details = {
                "path": self.request.path,
                "method": self.request.method,
                "body": self.request.body,
                "arguments": self.request.arguments,
                "body_arguments": self.request.body_arguments,
                "headers": dict(self.request.headers.items()),
                "query": self.request.query,
                "query_arguments": self.request.query_arguments,
                "uri": self.request.uri,
                "cookies": dict(self.request.cookies.items()),
                "response": responses,
            }
            with open(config.get("_security_risk_full_debugging.file"), "a+") as f:
                f.write(json.dumps(request_details, reject_bytes=False))
        super(BaseHandler, self).on_finish()

    async def attempt_sso_authn(self) -> bool:
        """
        ConsoleMe's configuration allows authenticating users by user/password, SSO, or both.
        This function helps determine how ConsoleMe should authenticate a user. If user/password login is allowed,
        users will be redirected to ConsoleMe's login page (/login). If SSO is also allowed, the Login page will present
        a button allowing the user to sign in with SSO.

        If user/password login is enabled, we don't want to give users the extra step of having to visit the login page,
        so we just authenticate them through SSO directly.
         allow authenticating users by a combination of user/password and SSO. In this case, we need to tell
        Returns: boolean
        """
        if not config.get("auth.get_user_by_password", False):
            return True

        # force_use_sso indicates the user's intent to authenticate via SSO
        force_use_sso = self.request.arguments.get("use_sso", [False])[0]
        if force_use_sso:
            return True
        # It's a redirect from an SSO provider. Let it hit the SSO functionality
        if (
            "code" in self.request.query_arguments
            and "state" in self.request.query_arguments
        ):
            return True
        if self.request.path == "/saml/acs":
            return True
        return False

    async def authorization_flow(
        self, user: str = None, console_only: bool = True, refresh_cache: bool = False
    ) -> None:
        """Perform high level authorization flow."""
        self.eligible_roles = []
        self.eligible_accounts = []
        self.request_uuid = str(uuid.uuid4())
        refresh_cache = (
            self.request.arguments.get("refresh_cache", [False])[0] or refresh_cache
        )
        attempt_sso_authn = await self.attempt_sso_authn()

        refreshed_user_roles_from_cache = False

        if not refresh_cache and config.get(
            "dynamic_config.role_cache.always_refresh_roles_cache", False
        ):
            refresh_cache = True

        self.red = await RedisHandler().redis()
        self.ip = self.get_request_ip()
        self.user = user
        self.groups = None
        self.user_role_name = None
        self.auth_cookie_expiration = 0

        log_data = {
            "function": "Basehandler.authorization_flow",
            "ip": self.ip,
            "request_path": self.request.uri,
            "user-agent": self.request.headers.get("User-Agent"),
            "request_id": self.request_uuid,
            "message": "Incoming request",
        }

        log.debug(log_data)

        # Check to see if user has a valid auth cookie
        if config.get("auth_cookie_name", "consoleme_auth"):
            auth_cookie = self.get_cookie(
                config.get("auth_cookie_name", "consoleme_auth")
            )

            # Validate auth cookie and use it to retrieve group information
            if auth_cookie:
                res = await validate_and_return_jwt_token(auth_cookie)
                if res and isinstance(res, dict):
                    self.user = res.get("user")
                    self.groups = res.get("groups")
                    self.auth_cookie_expiration = res.get("exp")

        if not self.user:
            # Check for development mode and a configuration override that specify the user and their groups.
            if config.get("development") and config.get("_development_user_override"):
                self.user = config.get("_development_user_override")
            if config.get("development") and config.get("_development_groups_override"):
                self.groups = config.get("_development_groups_override")

        if not self.user:
            # SAML flow. If user has a JWT signed by ConsoleMe, and SAML is enabled in configuration, user will go
            # through this flow.

            if config.get("auth.get_user_by_saml", False) and attempt_sso_authn:
                res = await authenticate_user_by_saml(self)
                if not res:
                    if (
                        self.request.uri != "/saml/acs"
                        and not self.request.uri.startswith("/auth?")
                    ):
                        raise SilentException(
                            "Unable to authenticate the user by SAML. "
                            "Redirecting to authentication endpoint"
                        )
                    return

        if not self.user:
            if config.get("auth.get_user_by_oidc", False) and attempt_sso_authn:
                res = await authenticate_user_by_oidc(self)
                if not res:
                    raise SilentException(
                        "Unable to authenticate the user by OIDC. "
                        "Redirecting to authentication endpoint"
                    )
                if res and isinstance(res, dict):
                    self.user = res.get("user")
                    self.groups = res.get("groups")

        if not self.user:
            if config.get("auth.get_user_by_aws_alb_auth", False):
                res = await authenticate_user_by_alb_auth(self)
                if not res:
                    raise Exception("Unable to authenticate the user by ALB Auth")
                if res and isinstance(res, dict):
                    self.user = res.get("user")
                    self.groups = res.get("groups")

        if not self.user:
            # Username/Password authn flow
            if config.get("auth.get_user_by_password", False):
                after_redirect_uri = self.request.arguments.get("redirect_url", [""])[0]
                if after_redirect_uri and isinstance(after_redirect_uri, bytes):
                    after_redirect_uri = after_redirect_uri.decode("utf-8")
                self.set_status(403)
                self.write(
                    {
                        "type": "redirect",
                        "redirect_url": f"/login?redirect_after_auth={after_redirect_uri}",
                        "reason": "unauthenticated",
                        "message": "User is not authenticated. Redirect to authenticate",
                    }
                )
                await self.finish()
                raise SilentException(
                    "Redirecting user to authenticate by username/password."
                )

        if not self.user:
            try:
                # Get user. Config options can specify getting username from headers or
                # OIDC, but custom plugins are also allowed to override this.
                self.user = await auth.get_user(headers=self.request.headers)
                if not self.user:
                    raise NoUserException(
                        f"User not detected. Headers: {self.request.headers}"
                    )
                log_data["user"] = self.user
            except NoUserException:
                self.clear()
                self.set_status(403)

                stats.count(
                    "Basehandler.authorization_flow.no_user_detected",
                    tags={
                        "request_path": self.request.uri,
                        "ip": self.ip,
                        "user_agent": self.request.headers.get("User-Agent"),
                    },
                )
                log_data["message"] = "No user detected. Check configuration."
                log.error(log_data)
                await self.finish(log_data["message"])
                raise

        self.contractor = config.config_plugin().is_contractor(self.user)

        if config.get("auth.cache_user_info_server_side", True) and not refresh_cache:
            try:
                cache_r = self.red.get(f"USER-{self.user}-CONSOLE-{console_only}")
            except redis.exceptions.ConnectionError:
                cache_r = None
            if cache_r:
                log_data["message"] = "Loading from cache"
                log.debug(log_data)
                cache = json.loads(cache_r)
                self.groups = cache.get("groups")
                self.eligible_roles = cache.get("eligible_roles")
                self.eligible_accounts = cache.get("eligible_accounts")
                self.user_role_name = cache.get("user_role_name")
                refreshed_user_roles_from_cache = True

        try:
            if not self.groups:
                self.groups = await auth.get_groups(
                    self.user, headers=self.request.headers
                )
            if not self.groups:
                raise NoGroupsException(
                    f"Groups not detected. Headers: {self.request.headers}"
                )

        except NoGroupsException:
            stats.count("Basehandler.authorization_flow.no_groups_detected")
            log_data["message"] = "No groups detected. Check configuration."
            log.error(log_data)

        # Set Per-User Role Name (This logic is not used in OSS deployment)
        if (
            config.get("user_roles.opt_in_group")
            and config.get("user_roles.opt_in_group") in self.groups
        ):
            # Get or create user_role_name attribute
            self.user_role_name = await auth.get_or_create_user_role_name(self.user)

        self.eligible_roles = await group_mapping.get_eligible_roles(
            self.user, self.groups, self.user_role_name, console_only=console_only
        )

        if not self.eligible_roles:
            log_data[
                "message"
            ] = "No eligible roles detected for user. But letting them continue"
            log.warning(log_data)
        log_data["eligible_roles"] = len(self.eligible_roles)

        if not self.eligible_accounts:
            try:
                self.eligible_accounts = await group_mapping.get_eligible_accounts(
                    self.eligible_roles
                )
                log_data["eligible_accounts"] = len(self.eligible_accounts)
                log_data["message"] = "Successfully authorized user."
                log.debug(log_data)
            except Exception:
                stats.count("Basehandler.authorization_flow.exception")
                log.error(log_data, exc_info=True)
                raise
        if (
            config.get("auth.cache_user_info_server_side", True)
            and self.groups
            # Only set role cache if we didn't retrieve user's existing roles from cache
            and not refreshed_user_roles_from_cache
        ):
            try:
                self.red.setex(
                    f"USER-{self.user}-CONSOLE-{console_only}",
                    config.get("dynamic_config.role_cache.cache_expiration", 60),
                    json.dumps(
                        {
                            "groups": self.groups,
                            "eligible_roles": self.eligible_roles,
                            "eligible_accounts": self.eligible_accounts,
                            "user_role_name": self.user_role_name,
                        }
                    ),
                )
            except redis.exceptions.ConnectionError:
                pass
        if (
            config.get("auth.set_auth_cookie")
            and config.get("auth_cookie_name", "consoleme_auth")
            and not self.get_cookie(config.get("auth_cookie_name", "consoleme_auth"))
        ):
            expiration = datetime.utcnow().replace(tzinfo=pytz.UTC) + timedelta(
                minutes=config.get("jwt.expiration_minutes", 60)
            )

            encoded_cookie = await generate_jwt_token(
                self.user, self.groups, exp=expiration
            )
            self.set_cookie(
                config.get("auth_cookie_name", "consoleme_auth"),
                encoded_cookie,
                expires=expiration,
                secure=config.get(
                    "auth.cookie.secure",
                    "https://" in config.get("url"),
                ),
                httponly=config.get("auth.cookie.httponly", True),
                samesite=config.get("auth.cookie.samesite", True),
            )
        if self.tracer:
            await self.tracer.set_additional_tags({"USER": self.user})


class BaseAPIV1Handler(BaseHandler):
    """Default API Handler for api/v1/* routes."""

    def set_default_headers(self) -> None:
        self.set_header("Content-Type", "application/json")


class BaseAPIV2Handler(BaseHandler):
    """Default API Handler for api/v2/* routes."""

    def set_default_headers(self) -> None:
        self.set_header("Content-Type", "application/json")

    def write_error(self, status_code: int, **kwargs: Any) -> None:
        if self.settings.get("serve_traceback") and "exc_info" in kwargs:
            # in debug mode, try to send a traceback
            self.set_header("Content-Type", "text/plain")
            self.set_status(status_code)
            for line in traceback.format_exception(*kwargs["exc_info"]):
                self.write(line)
            self.finish()
        else:
            self.set_header("Content-Type", "application/problem+json")
            title = httputil.responses.get(status_code, "Unknown")
            message = kwargs.get("message", self._reason)
            # self.set_status() modifies self._reason, so this call should come after we grab the reason
            self.set_status(status_code)
            self.finish(
                json.dumps(
                    {"status": status_code, "title": title, "message": message}
                )  # noqa
            )


class BaseMtlsHandler(BaseAPIV2Handler):
    def initialize(self, **kwargs):
        self.kwargs = kwargs

    async def prepare(self):
        self.tracer = None
        self.span = None
        self.spans = {}
        self.responses = []
        self.request_uuid = str(uuid.uuid4())
        self.auth_cookie_expiration = 0
        stats.timer("base_handler.incoming_request")
        if config.get("auth.require_mtls", False):
            try:
                await auth.validate_certificate(self.request.headers)
            except InvalidCertificateException:
                stats.count(
                    "GetCredentialsHandler.post.invalid_certificate_header_value"
                )
                self.set_status(403)
                self.write({"code": "403", "message": "Invalid Certificate"})
                await self.finish()
                return

            # Extract user from valid certificate
            try:
                self.requester = await auth.extract_user_from_certificate(
                    self.request.headers
                )
                self.current_cert_age = await auth.get_cert_age_seconds(
                    self.request.headers
                )
            except (MissingCertificateException, Exception) as e:
                if isinstance(e, MissingCertificateException):
                    stats.count("BaseMtlsHandler.post.missing_certificate_header")
                    message = "Missing Certificate in Header."
                else:
                    stats.count("BaseMtlsHandler.post.exception")
                    message = f"Invalid Mtls Certificate: {e}"
                self.set_status(400)
                self.write({"code": "400", "message": message})
                await self.finish()
                return
        elif config.get("auth.require_jwt", True):
            # Check to see if user has a valid auth cookie
            if config.get("auth_cookie_name", "consoleme_auth"):
                auth_cookie = self.get_cookie(
                    config.get("auth_cookie_name", "consoleme_auth")
                )

                if auth_cookie:
                    res = await validate_and_return_jwt_token(auth_cookie)
                    if not res:
                        error = {
                            "code": "invalid_jwt",
                            "message": "JWT is invalid or has expired.",
                            "request_id": self.request_uuid,
                        }
                        self.set_status(403)
                        self.write(error)
                        await self.finish()
                    self.user = res.get("user")
                    self.groups = res.get("groups")
                    self.requester = {"type": "user", "email": self.user}
                    self.current_cert_age = int(time.time()) - res.get("iat")
                    self.auth_cookie_expiration = res.get("exp")
            else:
                raise MissingConfigurationValue(
                    "Auth cookie name is not defined in configuration."
                )
        else:
            raise MissingConfigurationValue("Unsupported authentication scheme.")
        if not hasattr(self, "requester"):
            raise tornado.web.HTTPError(403, "Unable to authenticate user.")
        self.ip = self.get_request_ip()
        await self.configure_tracing()

    def write(self, chunk: Union[str, bytes, dict]) -> None:
        if config.get("_security_risk_full_debugging.enabled"):
            self.responses.append(chunk)
        super(BaseMtlsHandler, self).write(chunk)

    def on_finish(self) -> None:
        if config.get("_security_risk_full_debugging.enabled"):
            request_details = {
                "path": self.request.path,
                "method": self.request.method,
                "body": self.request.body,
                "arguments": self.request.arguments,
                "body_arguments": self.request.body_arguments,
                "headers": dict(self.request.headers.items()),
                "query": self.request.query,
                "query_arguments": self.request.query_arguments,
                "uri": self.request.uri,
                "cookies": dict(self.request.cookies.items()),
                "response": self.responses,
            }
            with open(config.get("_security_risk_full_debugging.file"), "a+") as f:
                f.write(json.dumps(request_details, reject_bytes=False))
        super(BaseMtlsHandler, self).on_finish()


class NoCacheStaticFileHandler(tornado.web.StaticFileHandler):
    def set_default_headers(self) -> None:
        self.set_header(
            "Cache-Control", "no-store, no-cache, must-revalidate, max-age=0"
        )



import time

from consoleme.handlers.base import BaseHandler


class AuthHandler(BaseHandler):
    async def prepare(self):
        try:
            if self.request.method.lower() in ["options", "post"]:
                return
            await super(AuthHandler, self).prepare()
        except:  # noqa
            # NoUserException
            raise

    async def get(self):
        self.write(
            {
                "authCookieExpiration": self.auth_cookie_expiration,
                "currentServerTime": int(time.time()),
            }
        )

    async def post(self):
        self.write(
            {
                "authCookieExpiration": self.auth_cookie_expiration,
                "currentServerTime": int(time.time()),
            }
        )



import json as original_json
import sys
import time
from collections import defaultdict
from typing import Dict

import sentry_sdk
import ujson as json
from asgiref.sync import sync_to_async

from consoleme.config import config
from consoleme.lib.cache import (
    retrieve_json_data_from_redis_or_s3,
    store_json_results_in_redis_and_s3,
)
from consoleme.lib.dynamo import UserDynamoHandler
from consoleme.lib.json_encoder import SetEncoder
from consoleme.lib.notifications.models import (
    ConsoleMeUserNotification,
    GetNotificationsForUserResponse,
)
from consoleme.lib.singleton import Singleton

log = config.get_logger()


class RetrieveNotifications(metaclass=Singleton):
    def __init__(self):
        self.last_update = 0
        self.all_notifications = []

    async def retrieve_all_notifications(self, force_refresh=False):
        if force_refresh or (
            int(time.time()) - self.last_update
            > config.get(
                "get_notifications_for_user.notification_retrieval_interval", 20
            )
        ):
            self.all_notifications = await retrieve_json_data_from_redis_or_s3(
                redis_key=config.get("notifications.redis_key", "ALL_NOTIFICATIONS"),
                redis_data_type="hash",
                s3_bucket=config.get("notifications.s3.bucket"),
                s3_key=config.get(
                    "notifications.s3.key", "notifications/all_notifications_v1.json.gz"
                ),
                default={},
            )
            self.last_update = int(time.time())
        return self.all_notifications


async def get_notifications_for_user(
    user,
    groups,
    max_notifications=config.get("get_notifications_for_user.max_notifications", 5),
    force_refresh=False,
) -> GetNotificationsForUserResponse:
    function = f"{__name__}.{sys._getframe().f_code.co_name}"
    log_data = {
        "function": function,
        "user": user,
        "max_notifications": max_notifications,
        "force_refresh": force_refresh,
    }

    current_time = int(time.time())
    all_notifications = await RetrieveNotifications().retrieve_all_notifications(
        force_refresh
    )
    unread_count = 0
    notifications_for_user = []
    for user_or_group in [user, *groups]:
        # Filter out identical notifications that were already captured via user-specific attribution. IE: "UserA"
        # performed an access deny operation locally under "RoleA" with session name = "UserA", so the generated
        # notification is tied to the user. However, "UserA" is a member of "GroupA", which owns RoleA. We want
        # to show the notification to members of "GroupA", as well as "UserA" but we don't want "UserA" to see 2
        # notifications.
        notifications = all_notifications.get(user_or_group)
        if not notifications:
            continue
        notifications = json.loads(notifications)
        for notification_raw in notifications:
            try:
                # We parse ConsoleMeUserNotification individually instead of as an array
                # to account for future changes to the model that may invalidate older
                # notifications
                notification = ConsoleMeUserNotification.parse_obj(notification_raw)
            except Exception as e:
                log.error({**log_data, "error": str(e)})
                sentry_sdk.capture_exception()
                continue
            if notification.version != 1:
                # Skip unsupported versions of the notification model
                continue
            if user in notification.hidden_for_users:
                # Skip this notification if it isn't hidden for the user
                continue
            seen = False
            for existing_user_notification_raw in notifications_for_user:
                existing_user_notification = ConsoleMeUserNotification.parse_obj(
                    existing_user_notification_raw
                )
                if (
                    notification.predictable_id
                    == existing_user_notification.predictable_id
                ):
                    seen = True
            if not seen:
                notifications_for_user.append(notification)
    # Filter out "expired" notifications
    notifications_for_user = [
        v for v in notifications_for_user if v.expiration > current_time
    ]

    # Show newest notifications first
    notifications_for_user = sorted(
        notifications_for_user, key=lambda i: i.event_time, reverse=True
    )

    # Increment Unread Count
    notifications_to_return = notifications_for_user[0:max_notifications]
    for notification in notifications_to_return:
        if user in notification.read_by_users or notification.read_by_all:
            notification.read_for_current_user = True
            continue
        unread_count += 1
    return GetNotificationsForUserResponse(
        notifications=notifications_to_return, unread_count=unread_count
    )


async def fetch_notification(notification_id: str):
    ddb = UserDynamoHandler()
    notification = await sync_to_async(ddb.notifications_table.get_item)(
        Key={"predictable_id": notification_id}
    )
    if notification.get("Item"):
        return ConsoleMeUserNotification.parse_obj(notification["Item"])


async def cache_notifications_to_redis_s3() -> Dict[str, int]:
    function = f"{__name__}.{sys._getframe().f_code.co_name}"
    current_time = int(time.time())
    log_data = {"function": function}
    ddb = UserDynamoHandler()
    notifications_by_user_group = defaultdict(list)
    all_notifications_l = await ddb.parallel_scan_table_async(ddb.notifications_table)
    changed_notifications = []
    for existing_notification in all_notifications_l:
        notification = ConsoleMeUserNotification.parse_obj(existing_notification)
        if current_time > notification.expiration:
            notification.expired = True
            changed_notifications.append(notification.dict())
        for user_or_group in notification.users_or_groups:
            notifications_by_user_group[user_or_group].append(notification.dict())

    if changed_notifications:
        ddb.parallel_write_table(ddb.notifications_table, changed_notifications)

    if notifications_by_user_group:
        for k, v in notifications_by_user_group.items():
            notifications_by_user_group[k] = original_json.dumps(v, cls=SetEncoder)
        await store_json_results_in_redis_and_s3(
            notifications_by_user_group,
            redis_key=config.get("notifications.redis_key", "ALL_NOTIFICATIONS"),
            redis_data_type="hash",
            s3_bucket=config.get("notifications.s3.bucket"),
            s3_key=config.get(
                "notifications.s3.key", "notifications/all_notifications_v1.json.gz"
            ),
        )
    log_data["num_user_groups_for_notifications"] = len(
        notifications_by_user_group.keys()
    )
    log_data["num_notifications"] = len(all_notifications_l)
    log.debug(log_data)
    return {
        "num_user_groups_to_notify": len(notifications_by_user_group.keys()),
        "num_notifications": len(all_notifications_l),
    }


async def write_notification(notification: ConsoleMeUserNotification):
    ddb = UserDynamoHandler()
    await sync_to_async(ddb.notifications_table.put_item)(
        Item=ddb._data_to_dynamo_replace(notification.dict())
    )
    await cache_notifications_to_redis_s3()
    return True



"""Configuration handling library."""
import collections
import datetime
import logging
import os
import socket
import sys
import threading
import time
from logging import LoggerAdapter, LogRecord
from threading import Timer
from typing import Any, Dict, List, Optional, Union

import boto3
import botocore.exceptions
import logmatic
import ujson as json
import yaml
from asgiref.sync import async_to_sync
from pytz import timezone

from consoleme.lib.aws_secret_manager import get_aws_secret
from consoleme.lib.plugins import get_plugin_by_name

config_plugin_entrypoint = os.environ.get(
    "CONSOLEME_CONFIG_ENTRYPOINT", "default_config"
)
config_plugin = get_plugin_by_name(config_plugin_entrypoint)
main_exit_flag = threading.Event()


def dict_merge(dct: dict, merge_dct: dict):
    """Recursively merge two dictionaries, including nested dicts"""
    for k, v in merge_dct.items():
        if (
            k in dct
            and isinstance(dct[k], dict)
            and isinstance(merge_dct[k], collections.Mapping)
        ):
            dict_merge(dct[k], merge_dct[k])
        else:
            # Prefer original dict values over merged dict values if they already exist
            if k not in dct.keys():
                dct[k] = merge_dct[k]
    return dct


def refresh_dynamic_config(ddb=None):
    if not ddb:
        # This function runs frequently. We provide the option to pass in a UserDynamoHandler
        # so we don't need to import on every invocation
        from consoleme.lib.dynamo import UserDynamoHandler

        ddb = UserDynamoHandler()
    return ddb.get_dynamic_config_dict()


class Configuration(object):
    """Load YAML configuration files. YAML files can be extended to extend each other, to include common configuration
    values."""

    def __init__(self) -> None:
        """Initialize empty configuration."""
        self.config = {}
        self.log = None

    def raise_if_invalid_aws_credentials(self):
        try:
            boto3.client(
                "sts", **self.get("boto3.client_kwargs", {})
            ).get_caller_identity()
        except botocore.exceptions.NoCredentialsError:
            raise Exception(
                "We were unable to detect valid AWS credentials. ConsoleMe needs valid AWS credentials to "
                "run.\n\n"
                "For local development: Provide credentials via environment variables, in your "
                "~/.aws/credentials file, or via Weep EC2 IMDS / ECS credential provider emulation.\n\n"
                "For a production configuration, please attach an IAM role to your instance(s) or container(s) through"
                "AWS.\n\n"
                "For more information, see how the Python AWS SDK retrieves credentials here: "
                "https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html#configuring-credentials"
            )

    def load_config_from_dynamo(self, ddb=None, red=None):
        if not ddb:
            from consoleme.lib.dynamo import UserDynamoHandler

            ddb = UserDynamoHandler()
        if not red:
            from consoleme.lib.redis import RedisHandler

            red = RedisHandler().redis_sync()

        dynamic_config = refresh_dynamic_config(ddb)
        if dynamic_config and dynamic_config != self.config.get("dynamic_config"):
            red.set(
                "DYNAMIC_CONFIG_CACHE",
                json.dumps(dynamic_config),
            )
            self.get_logger("config").debug(
                {
                    "function": f"{__name__}.{self.__class__.__name__}.{sys._getframe().f_code.co_name}",
                    "message": "Dynamic configuration changes detected and loaded",
                }
            )
            self.config["dynamic_config"] = dynamic_config

    def load_dynamic_config_from_redis(self, log_data: Dict[str, Any], red=None):
        if not red:
            from consoleme.lib.redis import RedisHandler

            red = RedisHandler().redis_sync()
        dynamic_config = red.get("DYNAMIC_CONFIG_CACHE")
        if not dynamic_config:
            self.get_logger("config").warning(
                {
                    **log_data,
                    "error": (
                        "Unable to retrieve Dynamic Config from Redis. "
                        "This can be safely ignored if your dynamic config is empty."
                    ),
                }
            )
            return
        dynamic_config_j = json.loads(dynamic_config)
        if self.config.get("dynamic_config", {}) != dynamic_config_j:
            self.get_logger("config").debug(
                {
                    **log_data,
                    "message": "Refreshing dynamic configuration from Redis",
                }
            )
            self.config["dynamic_config"] = dynamic_config_j

    def load_config_from_dynamo_bg_thread(self):
        """If enabled, we can load a configuration dynamically from Dynamo at a certain time interval. This reduces
        the need for code redeploys to make configuration changes"""
        from consoleme.lib.dynamo import UserDynamoHandler
        from consoleme.lib.redis import RedisHandler

        ddb = UserDynamoHandler()
        red = RedisHandler().redis_sync()

        while threading.main_thread().is_alive():
            self.load_config_from_dynamo(ddb, red)
            # Wait till main exit flag is set OR a fixed timeout
            if main_exit_flag.wait(
                timeout=self.get("dynamic_config.dynamo_load_interval", 60)
            ):
                break

    def __set_flag_on_main_exit(self):
        # while main thread is active, do nothing
        while threading.main_thread().is_alive():
            time.sleep(1)
        # Main thread exited, signal to other threads
        main_exit_flag.set()

    def purge_redislite_cache(self):
        """
        Purges redislite cache in primary DB periodically. This will force a cache refresh, and it is
        convenient for cases where you cannot securely run shared Redis (ie: AWS AppRunner)
        """
        if not self.get("redis.use_redislite"):
            return
        from consoleme.lib.redis import RedisHandler

        red = RedisHandler().redis_sync()
        while threading.main_thread().is_alive():
            red.flushdb()
            # Wait till main exit flag is set OR a fixed timeout
            if main_exit_flag.wait(
                timeout=self.get("redis.purge_redislite_cache_interval", 1800)
            ):
                break

    async def merge_extended_paths(self, extends, dir_path):
        for s in extends:
            extend_config = {}
            # This decode and YAML-load a string stored in AWS Secrets Manager
            if s.startswith("AWS_SECRETS_MANAGER:"):
                secret_name = "".join(s.split("AWS_SECRETS_MANAGER:")[1:])
                extend_config = yaml.safe_load(
                    get_aws_secret(
                        secret_name, os.environ.get("EC2_REGION", "us-east-1")
                    )
                )
            else:
                try:
                    extend_path = os.path.join(dir_path, s)
                    with open(extend_path, "r") as ymlfile:
                        extend_config = yaml.safe_load(ymlfile)
                except FileNotFoundError:
                    logging.error(f"Unable to open file: {s}", exc_info=True)

            dict_merge(self.config, extend_config)
            if extend_config.get("extends"):
                await self.merge_extended_paths(extend_config.get("extends"), dir_path)

    def reload_config(self):
        # We don't want to start additional background threads when we're reloading static configuration.
        while threading.main_thread().is_alive():
            async_to_sync(self.load_config)(
                allow_automatically_reload_configuration=False,
                allow_start_background_threads=False,
            )
            if not self.get("config.automatically_reload_configuration"):
                break
            # Wait till main exit flag is set OR a fixed timeout
            if main_exit_flag.wait(
                timeout=self.get("dynamic_config.reload_static_config_interval", 60)
            ):
                break

    async def load_config(
        self,
        allow_automatically_reload_configuration=True,
        allow_start_background_threads=True,
    ):
        """Load configuration from the location given to us by config_plugin"""
        path = config_plugin.get_config_location()

        try:
            with open(path, "r") as ymlfile:
                self.config = yaml.safe_load(ymlfile)
        except FileNotFoundError as e:
            raise FileNotFoundError(
                "File not found. Please set the CONFIG_LOCATION environmental variable "
                f"to point to ConsoleMe's YAML configuration file: {e}"
            )

        extends = self.get("extends")
        dir_path = os.path.dirname(path)

        if extends:
            await self.merge_extended_paths(extends, dir_path)

        if self.config.get("environment") != "test":
            self.raise_if_invalid_aws_credentials()

        # We use different Timer intervals for our background threads to prevent logger objects from clashing, which
        # could cause duplicate log entries.
        if allow_start_background_threads:
            Timer(0, self.__set_flag_on_main_exit, ()).start()

        if allow_start_background_threads and self.get("redis.use_redislite"):
            t = Timer(1, self.purge_redislite_cache, ())
            t.start()

        if allow_start_background_threads and self.get("config.load_from_dynamo", True):
            t = Timer(2, self.load_config_from_dynamo_bg_thread, ())
            t.start()

        if allow_start_background_threads and self.get(
            "config.run_recurring_internal_tasks"
        ):
            t = Timer(3, config_plugin.internal_functions, kwargs={"cfg": self.config})
            t.start()

        if allow_automatically_reload_configuration and self.get(
            "config.automatically_reload_configuration"
        ):
            t = Timer(4, self.reload_config, ())
            t.start()

    def get(
        self, key: str, default: Optional[Union[List[str], int, bool, str, Dict]] = None
    ) -> Any:
        """Get value for configuration entry in dot notation."""
        value = self.config
        for k in key.split("."):
            try:
                value = value[k]
            except KeyError:
                return default
        return value

    def get_logger(self, name: Optional[str] = None) -> LoggerAdapter:
        """Get logger."""
        if self.log:
            return self.log
        if not name:
            name = self.get("application_name", "consoleme")
        level_c = self.get("logging.level", "debug")
        if level_c == "info":
            level = logging.INFO
        elif level_c == "critical":
            level = logging.CRITICAL
        elif level_c == "error":
            level = logging.ERROR
        elif level_c == "warning":
            level = logging.WARNING
        elif level_c == "debug":
            level = logging.DEBUG
        else:
            # default
            level = logging.DEBUG
        filter_c = ContextFilter()
        format_c = self.get(
            "logging.format",
            "%(asctime)s - %(levelname)s - %(name)s - [%(filename)s:%(lineno)s - %(funcName)s() ] - %(message)s",
        )

        logging.basicConfig(level=level, format=format_c)
        logger = logging.getLogger(name)
        logger.addFilter(filter_c)

        extra = {"eventTime": datetime.datetime.now(timezone("US/Pacific")).isoformat()}

        now = datetime.datetime.now()

        # Elasticsearch logging
        if self.get("logging.elasticsearch_enabled", False):
            try:
                es = f"{self.get('logging.elasticsearch.host')}:{self.get('logging.elasticsearch.port')}"
                index_name = (
                    f"{self.get('logging.elasticsearch.index_name', 'consoleme')}-'"
                    f"{now.year}{now.month}{now.day}"
                )
                from consoleme.lib.elasticsearch import ESHandler

                handler = ESHandler(es, index_name)
                handler.setFormatter(
                    logmatic.JsonFormatter(
                        json_indent=self.get("logging.json_formatter.indent")
                    )
                )
                handler.setLevel(self.get("logging.elasticsearch.level", "INFO"))
                logger.addHandler(handler)
            except Exception:
                logger.error(
                    "Unable to configure Elasticsearch logging.", exc_info=True
                )
        # Log to stdout and disk
        if self.get("logging.stdout_enabled", True):
            logger.propagate = False
            handler = logging.StreamHandler(sys.stdout)
            handler.setFormatter(
                logmatic.JsonFormatter(
                    json_indent=self.get("logging.json_formatter.indent")
                )
            )
            handler.setLevel(self.get("logging.stdout.level", "DEBUG"))
            logger.addHandler(handler)
            logging_file = self.get("logging.file")
            if logging_file:
                if "~" in logging_file:
                    logging_file = os.path.expanduser(logging_file)
                os.makedirs(os.path.dirname(logging_file), exist_ok=True)
                file_handler = logging.FileHandler(logging_file)
                file_handler.setFormatter(
                    logmatic.JsonFormatter(
                        json_indent=self.get("logging.json_formatter.indent")
                    )
                )
                logger.addHandler(file_handler)
        self.log = logging.LoggerAdapter(logger, extra)
        return self.log

    def set_logging_levels(self):
        default_logging_levels = {
            "asyncio": "WARNING",
            "boto3": "CRITICAL",
            "boto": "CRITICAL",
            "botocore": "CRITICAL",
            "elasticsearch.trace": "ERROR",
            "elasticsearch": "ERROR",
            "nose": "CRITICAL",
            "parso.python.diff": "WARNING",
            "raven.base.client": "WARNING",
            "role_protect_client": "WARNING",
            "s3transfer": "CRITICAL",
            "spectator.HttpClient": "WARNING",
            "spectator.Registry": "WARNING",
            "urllib3": "ERROR",
            "redislite.client": "WARNING",
            "redislite.configuration": "WARNING",
        }
        for logger, level in self.get("logging_levels", default_logging_levels).items():
            logging.getLogger(logger).setLevel(level)

    def get_aws_region(self):
        region_checks = [
            # check if set through ENV vars
            os.environ.get("EC2_REGION"),
            os.environ.get("AWS_REGION"),
            os.environ.get("AWS_DEFAULT_REGION"),
            # else check if set in config or in boto already
            boto3.DEFAULT_SESSION.region_name if boto3.DEFAULT_SESSION else None,
            boto3.Session().region_name,
            boto3.client("s3", **self.get("boto3.client_kwargs", {})).meta.region_name,
            "us-east-1",
        ]
        for region in region_checks:
            if region:
                return region


class ContextFilter(logging.Filter):
    """Logging Filter for adding hostname to log entries."""

    hostname = socket.gethostname()

    def filter(self, record: LogRecord) -> bool:
        record.hostname = ContextFilter.hostname
        return True


CONFIG = Configuration()
async_to_sync(CONFIG.load_config)()

get = CONFIG.get
get_logger = CONFIG.get_logger

# Set logging levels
CONFIG.set_logging_levels()

values = CONFIG.config
region = CONFIG.get_aws_region()
hostname = socket.gethostname()
api_spec = {}
dir_ref = dir



#!/usr/bin/env python3

"""
CDK Application for running ConsoleMe on ECS
"""

import os

import boto3
import yaml
from aws_cdk import core as cdk

from consoleme_ecs_cdk.service.consoleme_ecs_service_stack import (
    ConsolemeEcsServiceStack,
)
from consoleme_ecs_cdk.service.consoleme_spoke_accounts_stack import (
    ConsolemeSpokeAccountsStack,
)
from consoleme_ecs_cdk.service.constants import BASE_NAME, SPOKE_BASE_NAME

config_yaml = yaml.load(open("config.yaml"), Loader=yaml.FullLoader)

spoke_accounts = config_yaml.get("spoke_accounts", [])

main_environment = cdk.Environment(
    account=boto3.client("sts").get_caller_identity().get("Account"),
    region=os.getenv("AWS_REGION"),
)

app = cdk.App()

for spoke_account_id in spoke_accounts:
    spoke_environment = cdk.Environment(
        account=spoke_account_id, region=os.getenv("AWS_REGION")
    )
    spoke_stack = ConsolemeSpokeAccountsStack(
        app, SPOKE_BASE_NAME, env=spoke_environment
    )  # Spoke account stack

consoleme_ecs_service_stack = ConsolemeEcsServiceStack(
    app, BASE_NAME, env=main_environment
)  # ConsoleMe account

app.synth()
